"""Placeholder."""
import collections

import numpy as np

import histomicstk.utils as utils
from histomicstk.preprocessing import color_conversion
from histomicstk.preprocessing.color_deconvolution.find_stain_index import \
    find_stain_index
from histomicstk.preprocessing.color_deconvolution.rgb_separate_stains_macenko_pca import \
    rgb_separate_stains_macenko_pca
from histomicstk.preprocessing.color_deconvolution.rgb_separate_stains_xu_snmf import \
    rgb_separate_stains_xu_snmf
from histomicstk.preprocessing.color_deconvolution.stain_color_map import \
    stain_color_map

from ._linalg import normalize
from .complement_stain_matrix import complement_stain_matrix


def color_deconvolution(im_rgb, w, I_0=None):
    """Perform color deconvolution.

    The given RGB Image `I` is first first transformed into optical density
    space, and then projected onto the stain vectors in the columns of the
    3x3 stain matrix `W`.

    For deconvolving H&E stained image use:

    `w` = array([[0.650, 0.072, 0], [0.704, 0.990, 0], [0.286, 0.105, 0]])

    Parameters
    ----------
    im_rgb : array_like
        Input RGB Image that needs to be deconvolved.
    w : array_like
        A 3x3 matrix containing the color vectors in columns.
        For two stain images the third column is zero and will be
        complemented using cross-product. At least two of the three
        columns must be non-zero.
    I_0 : float or array_like, optional
        A float a 3-vector containing background RGB intensities.
        If unspecified, use the old OD conversion.

    Returns
    -------
    Stains : array_like
        An rgb image where in each channel contains the image of the
        stain of the corresponding column in the stain matrix `W`.
        The intensity range of each channel is [0, 255] suitable for
        displaying.
    StainsFloat : array_like
        An intensity image of deconvolved stains that is unbounded,
        suitable for reconstructing color images of deconvolved stains
        with color_convolution.
    Wc : array_like
        A 3x3 complemented stain matrix. Useful for color image
        reconstruction with color_convolution.

    See Also
    --------
    histomicstk.preprocessing.color_deconvolution.complement_stain_matrix,
    histomicstk.preprocessing.color_deconvolution.color_convolution
    histomicstk.preprocessing.color_conversion.rgb_to_od
    histomicstk.preprocessing.color_conversion.od_to_rgb
    histomicstk.preprocessing.color_conversion.rgb_to_sda
    histomicstk.preprocessing.color_conversion.sda_to_rgb

    """
    # complement stain matrix if needed
    w = np.array(w)
    if w.shape[1] < 3:
        wc = np.zeros((w.shape[0], 3))
        wc[:, :w.shape[1]] = w
        w = wc

    if np.linalg.norm(w[:, 2]) <= 1e-16:
        wc = complement_stain_matrix(w)
    else:
        wc = w

    # normalize stains to unit-norm
    wc = normalize(wc)

    # invert stain matrix
    Q = np.linalg.pinv(wc)

    # transform 3D input image to 2D RGB matrix format
    m = utils.convert_image_to_matrix(im_rgb)[:3]

    # transform input RGB to optical density values and deconvolve,
    # tfm back to RGB
    sda_fwd = color_conversion.rgb_to_sda(m, I_0)
    sda_deconv = np.dot(Q, sda_fwd)
    sda_inv = color_conversion.sda_to_rgb(sda_deconv,
                                          255 if I_0 is not None else None)

    # reshape output
    StainsFloat = utils.convert_matrix_to_image(sda_inv, im_rgb.shape)

    # transform type
    Stains = StainsFloat.clip(0, 255).astype(np.uint8)

    # return
    Unmixed = collections.namedtuple('Unmixed',
                                     ['Stains', 'StainsFloat', 'Wc'])
    Output = Unmixed(Stains, StainsFloat, wc)

    return Output


def _reorder_stains(W, stains=None):
    """Reorder stains in a stain matrix to a specific order.

    This is particularly relevant in macenco where the order of stains is not
    preserved during stain unmixing, so this method uses
    histomicstk.preprocessing.color_deconvolution.find_stain_index
    to reorder the stains matrix to the order provided by this parameter

    Parameters
    ----------
    W : np array
        A 3x3 matrix of stain column vectors.
    stains : list, optional
        List of stain names (order is important). Default is H&E.

    Returns
    -------
    np array
        A re-ordered 3x3 matrix of stain column vectors.

    """
    stains = ['hematoxylin', 'eosin'] if stains is None else stains

    assert len(stains) == 2, 'Only two-stain matrices are supported for now.'

    def _get_channel_order(W):
        first = find_stain_index(stain_color_map[stains[0]], W)
        second = 1 - first
        # If 2 stains, third "stain" is cross product of 1st 2 channels
        # calculated using complement_stain_matrix()
        third = 2
        return first, second, third

    def _ordered_stack(mat, order):
        return np.stack([mat[..., j] for j in order], -1)

    return _ordered_stack(W, _get_channel_order(W))


def stain_unmixing_routine(
        im_rgb, stains=None, stain_unmixing_method='macenko_pca',
        stain_unmixing_params=None, mask_out=None):
    """Perform stain unmixing using the method of choice (wrapper).

    Parameters
    ----------
    im_rgb : array_like
        An RGB image (m x n x 3) to unmix.

    stains : list, optional
        List of stain names (order is important). Default is H&E. This is
        particularly relevant in macenco where the order of stains is not
        preserved during stain unmixing, so this method uses
        histomicstk.preprocessing.color_deconvolution.find_stain_index
        to reorder the stains matrix to the order provided by this parameter

    stain_unmixing_method : str, default is 'macenko_pca'
        stain unmixing method to use. It should be one of the following
        'macenko_pca', or 'xu_snmf'.

    stain_unmixing_params : dict, default is an empty dict
        kwargs to pass as-is to the stain unmixing method.

    mask_out : array_like, default is None
        if not None, should be (m x n) boolean numpy array.
        This parameter ensures exclusion of non-masked areas from calculations
        and normalization. This is relevant because elements like blood,
        sharpie marker, white space, etc may throw off the normalization.

    Returns
    -------
    Wc : array_like
        A 3x3 complemented stain matrix.

    See Also
    --------
    histomicstk.preprocessing.color_deconvolution.separate_stains_macenko_pca
    histomicstk.preprocessing.color_deconvolution.separate_stains_xu_snmf

    References
    ----------
    .. [#] Macenko, M., Niethammer, M., Marron, J. S., Borland, D.,
           Woosley, J. T., Guan, X., ... & Thomas, N. E. (2009, June).
           A method for normalizing histology slides for quantitative analysis.
           In Biomedical Imaging: From Nano to Macro, 2009.  ISBI'09.
           IEEE International Symposium on (pp. 1107-1110). IEEE.
    .. [#] Xu, J., Xiang, L., Wang, G., Ganesan, S., Feldman, M., Shih, N. N.,
           ...& Madabhushi, A. (2015). Sparse Non-negative Matrix Factorization
           (SNMF) based color unmixing for breast histopathological image
           analysis.  Computerized Medical Imaging and Graphics, 46, 20-29.

    """
    stains = ['hematoxylin', 'eosin'] if stains is None else stains
    stain_unmixing_params = {} if stain_unmixing_params is None else stain_unmixing_params

    stain_unmixing_method = stain_unmixing_method.lower()

    if stain_unmixing_method == 'macenko_pca':
        stain_deconvolution = rgb_separate_stains_macenko_pca
        stain_unmixing_params['I_0'] = None
        stain_unmixing_params['mask_out'] = mask_out

    elif stain_unmixing_method == 'xu_snmf':
        stain_deconvolution = rgb_separate_stains_xu_snmf
        stain_unmixing_params['I_0'] = None
        assert mask_out is None, 'Masking is not yet implemented in xu_snmf.'

    else:
        msg = 'Unknown/Unimplemented deconvolution method.'
        raise ValueError(msg)

    # get W_source
    W_source = stain_deconvolution(im_rgb, **stain_unmixing_params)

    # If Macenco method, reorder channels in W_target and W_source as desired.
    # This is actually a necessary step in macenko's method since we're
    # not guaranteed the order of the different stains.
    if stain_unmixing_method == 'macenko_pca':
        W_source = _reorder_stains(W_source, stains=stains)

    return W_source


def color_deconvolution_routine(
        im_rgb, W_source=None, mask_out=None, **kwargs):
    """Unmix stains mixing followed by deconvolution (wrapper).

    Parameters
    ----------
    im_rgb : array_like
        An RGB image (m x n x 3) to color normalize

    W_source : np array, default is None
        A 3x3 matrix of source stain column vectors. Only provide this
        if you know the stains matrix in advance (unlikely) and would
        like to perform supervised deconvolution. If this is not provided,
        stain_unmixing_routine() is used to estimate W_source.

    mask_out : array_like, default is None
        if not None, should be (m x n) boolean numpy array.
        This parameter ensures exclusion of non-masked areas from calculations
        and stain matrix. This is relevant because elements like blood,
        sharpie marker, white space, cannot be modeled as a mix of two stains.

    kwargs : k,v pairs
        Passed as-is to stain_unmixing_routine() if W_source is None.

    Returns
    -------
    Output from color_deconvolution()

    See Also
    --------
    histomicstk.preprocessing.color_deconvolution.stain_unmixing_routine
    histomicstk.preprocessing.color_deconvolution.color_deconvolution

    """
    # get W_source if not provided
    if W_source is None:
        W_source = stain_unmixing_routine(im_rgb, mask_out=mask_out, **kwargs)

    # deconvolve
    Stains, StainsFloat, wc = color_deconvolution(im_rgb, w=W_source, I_0=None)

    # mask out (keep in mind, image is inverted)
    if mask_out is not None:
        for i in range(3):
            Stains[..., i][mask_out] = 255
            StainsFloat[..., i][mask_out] = 255.

    return Stains, StainsFloat, wc



"""
This package contains utility functions to augment images.

This is relevant in training convolutional neural network models. Refer to this study ...

Tellez, David, Geert Litjens, Peter Bandi, Wouter Bulten, John-Melle Bokhorst,
Francesco Ciompi, and Jeroen van der Laak. "Quantifying the effects of data
augmentation and stain color normalization in convolutional neural networks
for computational pathology." arXiv preprint arXiv:1902.06543 (2019).
"""
# make functions available at the package level using these shadow imports
# since we mostly have one function per file
from .color_augmentation import (perturb_stain_concentration,
                                 rgb_perturb_stain_concentration)

# list out things that are available for public use
__all__ = (

    # functions and classes of this package
    'perturb_stain_concentration',
    'rgb_perturb_stain_concentration',
)



"""
Created on Mon Sep 23 21:17:43 2019.

@author: mtageld
"""
import numpy as np
from PIL import Image
from skimage.color import rgb2gray

from histomicstk.annotations_and_masks.annotation_and_mask_utils import \
    get_image_from_htk_response
from histomicstk.annotations_and_masks.masks_to_annotations_handler import (
    get_annotation_documents_from_contours, get_contours_from_mask)
from histomicstk.features.compute_haralick_features import \
    compute_haralick_features
from histomicstk.features.compute_intensity_features import \
    compute_intensity_features
from histomicstk.preprocessing.color_conversion import lab_mean_std
from histomicstk.preprocessing.color_normalization import reinhard
from histomicstk.saliency.tissue_detection import (
    _deconv_color, get_slide_thumbnail,
    get_tissue_boundary_annotation_documents, get_tissue_mask)
from histomicstk.utils.general_utils import Base_HTK_Class

Image.MAX_IMAGE_PIXELS = None


class CD_single_tissue_piece:
    """Detect cellular regions in a single tissue piece (internal)."""

    def __init__(self, cd, tissue_mask, monitorPrefix=''):
        """Detect cellularity in one tissue piece (Internal).

        Arguments:
        ---------
        cd : object
            Cellularity_detector_superpixels instance
        tissue_mask : np array
            (mxn) mask of the tissue piece at cd.MAG magnification
        monitorPrefix : str
            Text to prepend to printed statements

        """
        self.cd = cd
        self.tissue_mask = 0 + tissue_mask
        self.monitorPrefix = monitorPrefix

    def run(self):
        """Get cellularity and optionally visualize on DSA."""
        # get cellularity
        self.restrict_mask_to_single_tissue_piece()
        self.cd._print2('%s: set_tissue_rgb()' % self.monitorPrefix)
        self.set_tissue_rgb()
        self.cd._print2('%s: set_superpixel_mask()' % self.monitorPrefix)
        self.set_superpixel_mask()
        self.cd._print2('%s: set_superpixel_features()' % self.monitorPrefix)
        self.set_superpixel_features()
        self.cd._print2('%s: set_superpixel_assignment()' % self.monitorPrefix)
        self.set_superpixel_assignment()
        self.cd._print2('%s: assign_cellularity_scores()' % self.monitorPrefix)
        self.assign_cellularity_scores()

        # visualize
        if self.cd.visualize_spixels or self.cd.visualize_contiguous:
            self.assign_colors_to_spixel_clusters()

        if self.cd.visualize_spixels:
            self.cd._print2(
                '%s: visualize_individual_superpixels()' % self.monitorPrefix)
            self.visualize_individual_superpixels()

        if self.cd.visualize_contiguous:
            self.cd._print2(
                '%s: visualize_contiguous_superpixels()' % self.monitorPrefix)
            self.visualize_contiguous_superpixels()

    def restrict_mask_to_single_tissue_piece(self):
        """Only keep relevant part of slide mask."""
        # find coordinates at scan magnification
        tloc = np.argwhere(self.tissue_mask)
        F = self.cd.slide_info['F_tissue']
        self.ymin, self.xmin = (int(j) for j in np.min(tloc, axis=0) * F)
        self.ymax, self.xmax = (int(j) for j in np.max(tloc, axis=0) * F)
        self.tissue_mask = self.tissue_mask[
            int(self.ymin / F): int(self.ymax / F),
            int(self.xmin / F): int(self.xmax / F)]

    def set_tissue_rgb(self):
        """Load RGB from server for single tissue piece."""
        # load RGB for this tissue piece at saliency magnification
        getStr = '/item/%s/tiles/region?left=%d&right=%d&top=%d&bottom=%d&encoding=PNG' % (
            self.cd.slide_id, self.xmin, self.xmax, self.ymin, self.ymax,
        ) + '&magnification=%d' % self.cd.MAG
        resp = self.cd.gc.get(getStr, jsonResp=False)
        self.tissue_rgb = get_image_from_htk_response(resp)

        # color normalization if desired
        if 'main' in self.cd.cnorm_params.keys():
            self.tissue_rgb = np.uint8(reinhard(
                im_src=self.tissue_rgb,
                target_mu=self.cd.cnorm_params['main']['mu'],
                target_sigma=self.cd.cnorm_params['main']['sigma']))

    def set_superpixel_mask(self):
        """Use Simple Linear Iterative Clustering (SLIC) to get superpixels."""
        from skimage.segmentation import slic
        from skimage.transform import resize

        # Get superpixel size and number
        spixel_size = self.cd.spixel_size_baseMag * (
            self.cd.MAG / self.cd.slide_info['magnification'])
        n_spixels = int(
            self.tissue_rgb.shape[0] * self.tissue_rgb.shape[1] / spixel_size)

        # get superpixel mask
        # optionally use grayscale instead of RGB -- seems more robust to
        # color variations and sometimes gives better results
        if self.cd.use_grayscale:
            try:
                self.spixel_mask = slic(
                    rgb2gray(self.tissue_rgb), n_segments=n_spixels,
                    compactness=self.cd.compactness, channel_axis=None)
            except Exception:
                self.spixel_mask = slic(
                    rgb2gray(self.tissue_rgb), n_segments=n_spixels,
                    compactness=self.cd.compactness)
        else:
            self.spixel_mask = slic(
                self.tissue_rgb, n_segments=n_spixels,
                compactness=self.cd.compactness)

        # restrict to tissue mask
        tmask = resize(
            self.tissue_mask, output_shape=self.spixel_mask.shape,
            order=0, preserve_range=True, anti_aliasing=False)
        self.spixel_mask[tmask == 0] = 0

    def set_superpixel_features(self):
        """Get superpixel features."""
        from pandas import concat
        from skimage.measure import regionprops

        assert (self.cd.use_intensity or self.cd.use_texture)

        # Possibly deconvolvve to get hematoxylin channel (cellular areas)
        # hematoxylin channel return shows MINIMA so we invert
        if self.cd.deconvolve:
            Stains, channel = _deconv_color(self.tissue_rgb)
            tissue_htx = 255 - Stains[..., channel]
        else:
            tissue_htx = rgb2gray(self.tissue_rgb)

        # calculate features from superpixels
        rprops = regionprops(self.spixel_mask)
        fdata_list = []
        if self.cd.use_intensity:
            fdata_list.append(compute_intensity_features(
                im_label=self.spixel_mask, im_intensity=tissue_htx,
                rprops=rprops))
        if self.cd.use_texture:
            fdata_list.append(compute_haralick_features(
                im_label=self.spixel_mask, im_intensity=tissue_htx,
                rprops=rprops))
        self.fdata = concat(fdata_list, axis=1)

        if self.cd.keep_feats is not None:
            self.fdata = self.fdata.loc[:, self.cd.keep_feats]

        # Index is corresponding pixel value in the superpixel mask
        # IMPORTANT: this assumes that regionprops output is sorted by unique
        # pixel values in label mask, which it is by default
        self.fdata.index = set(np.unique(self.spixel_mask)) - {0}

    def set_superpixel_assignment(self):
        """Fit gaussian mixture model to features and get assignment."""
        from sklearn.mixture import GaussianMixture

        mmodel = GaussianMixture(n_components=self.cd.n_gaussian_components)
        self.spixel_labels = mmodel.fit_predict(self.fdata.values) + 1

    def assign_cellularity_scores(self):
        """Assign cellularity scores to spixel clusters."""
        assert self.cd.use_intensity, 'We need intensity to rank cellularity.'
        assert self.cd.deconvolve, \
            'We must use hematoxyling channel to rank by cellularity.'

        self.cluster_props = {}

        self.fdata.loc[:, 'cluster'] = self.spixel_labels
        for clid in np.unique(self.spixel_labels):
            self.cluster_props[clid] = {
                'cellularity': int(np.median(self.fdata.loc[
                    self.fdata.loc[:, 'cluster'] == clid,
                    'Intensity.Median']) / 255 * 100),
            }

    def assign_colors_to_spixel_clusters(self):
        """Assign RGB color string to cellularity clusters."""
        # normalize values by given value, else by max for each tissue piece
        if self.cd.max_cellularity is not None:
            max_cellularity = self.cd.max_cellularity
        else:
            max_cellularity = max(
                j['cellularity'] for _, j in self.cluster_props.items()
            )

        # Assign rgb string
        for clid in np.unique(self.spixel_labels):
            cellularity = min(
                self.cluster_props[clid]['cellularity'], max_cellularity)
            rgb = self.cd.cMap(int(cellularity / max_cellularity * 255))[:-1]
            rgb = [int(255 * j) for j in rgb]
            self.cluster_props[clid]['color'] = 'rgb(%d,%d,%d)' % tuple(rgb)

    def visualize_individual_superpixels(self):
        """Visualize individual spixels, color-coded by cellularity."""
        # Define GTCodes dataframe
        from pandas import DataFrame

        GTCodes_df = DataFrame(columns=['group', 'GT_code', 'color'])
        for spval, sp in self.fdata.iterrows():
            spstr = 'spixel-%d_cellularity-%d' % (
                spval, self.cluster_props[sp['cluster']]['cellularity'])
            GTCodes_df.loc[spstr, 'group'] = spstr
            GTCodes_df.loc[spstr, 'GT_code'] = spval
            GTCodes_df.loc[spstr, 'color'] = \
                self.cluster_props[sp['cluster']]['color']

        # get contours df
        contours_df = get_contours_from_mask(
            MASK=self.spixel_mask, GTCodes_df=GTCodes_df,
            get_roi_contour=False, MIN_SIZE=0, MAX_SIZE=None,
            verbose=self.cd.verbose == 3, monitorPrefix=self.monitorPrefix)
        contours_df.loc[:, 'group'] = [
            j.split('_')[-1] for j in contours_df.loc[:, 'group']]

        # get annotation docs
        annprops = {
            'F': (self.ymax - self.ymin) / self.tissue_rgb.shape[0],
            'X_OFFSET': self.xmin,
            'Y_OFFSET': self.ymin,
            'opacity': self.cd.opacity,
            'lineWidth': self.cd.lineWidth,
        }
        annotation_docs = get_annotation_documents_from_contours(
            contours_df.copy(), docnamePrefix='spixel', annprops=annprops,
            annots_per_doc=1000, separate_docs_by_group=True,
            verbose=self.cd.verbose == 3, monitorPrefix=self.monitorPrefix)
        for didx, doc in enumerate(annotation_docs):
            self.cd._print2('%s: Posting doc %d of %d' % (
                self.monitorPrefix, didx + 1, len(annotation_docs)))
            _ = self.cd.gc.post(
                '/annotation?itemId=' + self.cd.slide_id, json=doc)

    def visualize_contiguous_superpixels(self):
        """Visualize contiguous spixels, color-coded by cellularity."""
        from pandas import DataFrame

        # get cellularity cluster membership mask
        cellularity_mask = np.zeros(self.spixel_mask.shape)
        for spval, sp in self.fdata.iterrows():
            cellularity_mask[self.spixel_mask == spval] = sp['cluster']

        # Define GTCodes dataframe
        GTCodes_df = DataFrame(columns=['group', 'GT_code', 'color'])
        for spval, cp in self.cluster_props.items():
            spstr = 'cellularity-%d' % (cp['cellularity'])
            GTCodes_df.loc[spstr, 'group'] = spstr
            GTCodes_df.loc[spstr, 'GT_code'] = spval
            GTCodes_df.loc[spstr, 'color'] = cp['color']

        # get contours df
        contours_df = get_contours_from_mask(
            MASK=cellularity_mask, GTCodes_df=GTCodes_df,
            get_roi_contour=False, MIN_SIZE=0, MAX_SIZE=None,
            verbose=self.cd.verbose == 3, monitorPrefix=self.monitorPrefix)

        # get annotation docs
        annprops = {
            'F': (self.ymax - self.ymin) / self.tissue_rgb.shape[0],
            'X_OFFSET': self.xmin,
            'Y_OFFSET': self.ymin,
            'opacity': self.cd.opacity_contig,
            'lineWidth': self.cd.lineWidth,
        }
        annotation_docs = get_annotation_documents_from_contours(
            contours_df.copy(), docnamePrefix='contig', annprops=annprops,
            annots_per_doc=1000, separate_docs_by_group=True,
            verbose=self.cd.verbose == 3, monitorPrefix=self.monitorPrefix)
        for didx, doc in enumerate(annotation_docs):
            self.cd._print2('%s: Posting doc %d of %d' % (
                self.monitorPrefix, didx + 1, len(annotation_docs)))
            _ = self.cd.gc.post(
                '/annotation?itemId=' + self.cd.slide_id, json=doc)


class Cellularity_detector_superpixels (Base_HTK_Class):
    """Detect cellular regions in a slides by classifying superpixels.

    This uses Simple Linear Iterative Clustering (SLIC) to get superpixels at
    a low slide magnification to detect cellular regions. The first step of
    this pipeline detects tissue regions (i.e. individual tissue pieces)
    using the get_tissue_mask method of the histomicstk.saliency module. Then,
    each tissue piece is processed separately for accuracy and disk space
    efficiency. It is important to keep in mind that this does NOT rely on a
    tile iterator, but loads the entire tissue region (but NOT the whole slide)
    in memory and passes it on to skimage.segmentation.slic method.

    Once superpixels are segmented, the image is deconvolved and features are
    extracted from the hematoxylin channel. Features include intensity and
    possibly also texture features. Then, a mixed component Gaussian mixture
    model is fit to the features, and median intensity is used to rank
    superpixel clusters by 'cellularity' (since we are working with the
    hematoxylin channel).

    Additional functionality includes contour extraction to get the final
    segmentation boundaries of cellular regions and to visualize them in DSA
    using one's preferred colormap.

    """

    def __init__(self, gc, slide_id, **kwargs):
        """Init Cellularity_Detector_Superpixels object.

        Arguments:
        ---------
        gc : object
            girder client object
        slide_id : str
            girder ID of slide
        verbose : int
            0 - Do not print to screen
            1 - Print only key messages
            2 - Print everything to screen
            3 - print everything including from inner functions
        monitorPrefix : str
            text to prepend to printed statements
        logging_savepath : str or None
            where to save run logs
        suppress_warnings : bool
            whether to suppress warnings
        cnorm_params : dict
            Reinhard color normalization parameters. Accepted keys: thumbnail
            and main (since thumbnail normalization is different from color
            normalization of tissue at target magnification. Each entry is a
            dict containing values for mu and sigma. This is either given
            here or can be set using self.set_color_normalization_values().
            May be left unset if you do not want to normalize.
        get_tissue_mask_kwargs : dict
            kwargs for the get_tissue_mask() method.
        MAG : float
            magnification at which to detect cellularity
        spixel_size_baseMag : int
            approximate superpixel size at base (scan) magnification
        compactness : float
            compactness parameter for the SLIC method. Higher values result
            in more regular superpixels while smaller values are more likely
            to respect tissue boundaries.
        deconvolve : bool
            Whether to deconvolve and use hematoxylin channel for feature
            extraction. Must be True to ranks spixel clusters by cellularity.
        use_grayscale : bool
            If True, grayscale image is used with SLIC. May be more robust to
            color variations from slide to slide and more efficient.
        use_intensity : bool
            Whether to extract intensity features from the hematoxylin channel.
            This must be True to rank spuerpixel clusters by cellularity.
        use_texture : bool
            Whether to extract Haralick texture features from Htx channel. May
            not necessarily improve results when used in conjunction with
            intensity features.
        keep_feats : list
            Name of intensity features to use. See
            histomicstk.features.compute_intensity_features.
            Using fewer informative features may result in better
            gaussian mixture modeling results.
        n_gaussian_components : int
            no of gaussian mixture model components
        max_cellularity : int
            Range [0, 100] or None. If None, normalize visualization RGB values
            for each tissue piece separately, else normalize by given number.
        opacity : float
            opacity of superpixel polygons when posted to DSA.
            0 (no opacity) is more efficient to render.
        opacity_contig : float
            opacity of contiguous region polygons when posted to DSA.
            0 (no opacity) is more efficient to render.
        lineWidth : float
            width of line when displaying superpixel boundaries.
        cMap : object
            matplotlib color map to use when visualizing cellularity
        visualize_tissue_boundary : bool
            whether to visualize result from tissue detection component
        visualize_spixels : bool
            whether to visualize superpixels, color-coded by cellularity
        visualize_contiguous : bool
            whether to visualize contiguous cellular regions

        """
        from matplotlib import cm

        default_attr = {

            # The following are already assigned defaults by Base_HTK_Class
            # 'verbose': 1,
            # 'monitorPrefix': "",
            # 'logging_savepath': None,
            # 'suppress_warnings': False,

            'cnorm_params': {},
            'MAG': 3.0,
            'get_tissue_mask_kwargs': {
                'deconvolve_first': True, 'n_thresholding_steps': 1,
                'sigma': 1.5, 'min_size': 500,
            },
            'spixel_size_baseMag': 256 * 256,
            'compactness': 0.1,
            'deconvolve': True,
            'use_grayscale': True,
            'use_intensity': True,
            'use_texture': False,
            # 'keep_feats': None,  # keep everything
            'keep_feats': [
                'Intensity.Mean', 'Intensity.Median',
                'Intensity.Std', 'Intensity.IQR',
                'Intensity.HistEntropy',
            ],
            'n_gaussian_components': 5,
            'max_cellularity': None,
            'opacity': 0,
            'opacity_contig': 0.3,
            'lineWidth': 3.0,
            'cMap': cm.seismic,
            'visualize_tissue_boundary': True,
            'visualize_spixels': True,
            'visualize_contiguous': True,
        }
        default_attr.update(kwargs)
        super().__init__(
            default_attr=default_attr)

        # set attribs
        self.gc = gc
        self.slide_id = slide_id

    def run(self):
        """Run cellularity detection and optionally visualize result.

        This runs the cellularity detection +/- visualization pipeline and
        returns a list of CD_single_tissue_piece objects. Each object has
        the following attributes

        tissue_mask : np array
            mask of where tissue is at target magnification
        ymin : int
            min y coordinate at base (scan) magnification
        xmin : int
            min x coordinate at base (scan) magnification
        ymax : int
            max y coordinate at base (scan) magnification
        xmax : int
            max x coordinate at base (scan) magnification
        spixel_mask : np array
            np array where each unique value represents one superpixel
        fdata : pandas DataFrame
            features extracted for each superpixel. Index corresponds to
            values in the spixel_mask. This includes a 'cluster' column
            indicatign which cluster this superpixel belongs to.
        cluster_props : dict
            properties of each superpixel cluster, including its assigned
            cellularity score.

        """
        if (len(self.cnorm_params) == 0) and (not self.suppress_warnings):
            input("""
                %s: WARNING!! Consider running set_color_normalization_values()
                first, using what='thumbnail' and/or what='main' before running
                this method. Continue anyway?""" % self.monitorPrefix)

        # get mask, each unique value is a single tissue piece
        self._print1(
            '%s: set_slide_info_and_get_tissue_mask()' % self.monitorPrefix)
        labeled = self.set_slide_info_and_get_tissue_mask()

        # Go through tissue pieces and do run sequence
        unique_tvals = list(set(np.unique(labeled)) - {0})
        tissue_pieces = [None for _ in range(len(unique_tvals))]
        for idx, tval in enumerate(unique_tvals):
            monitorPrefix = '%s: Tissue piece %d of %d' % (
                self.monitorPrefix, idx + 1, len(unique_tvals))
            self._print1(monitorPrefix)
            tissue_pieces[idx] = CD_single_tissue_piece(
                self, tissue_mask=labeled == tval, monitorPrefix=monitorPrefix)
            tissue_pieces[idx].run()
            del tissue_pieces[idx].tissue_rgb  # too much space

        return tissue_pieces

    def set_color_normalization_values(
            self, mu=None, sigma=None, ref_image_path=None, what='main'):
        """Set color normalization values for thumbnail or main image."""
        assert (
            all(j is not None for j in (mu, sigma)) or ref_image_path is not None
        ), 'You must provide mu & sigma values or ref. image to get them.'

        assert what in ('thumbnail', 'main')

        if ref_image_path is not None:
            from imageio.v2 import imread

            ref_im = np.array(imread(ref_image_path, pilmode='RGB'))
            mu, sigma = lab_mean_std(ref_im)

        self.cnorm_params[what] = {'mu': mu, 'sigma': sigma}

    def set_slide_info_and_get_tissue_mask(self):
        """Set self.slide_info dict and self.labeled tissue mask."""
        # This is a persistent dict to store information about slide
        self.slide_info = self.gc.get('item/%s/tiles' % self.slide_id)

        # get tissue mask
        thumbnail_rgb = get_slide_thumbnail(self.gc, self.slide_id)

        # color normalization if desired
        if 'thumbnail' in self.cnorm_params.keys():
            thumbnail_rgb = np.uint8(reinhard(
                im_src=thumbnail_rgb,
                target_mu=self.cnorm_params['thumbnail']['mu'],
                target_sigma=self.cnorm_params['thumbnail']['sigma']))

        # get labeled tissue mask -- each unique value is one tissue piece
        labeled, _ = get_tissue_mask(
            thumbnail_rgb, **self.get_tissue_mask_kwargs)

        if len(np.unique(labeled)) < 2:
            msg = 'No tissue detected!'
            raise ValueError(msg)

        if self.visualize_tissue_boundary:
            annotation_docs = get_tissue_boundary_annotation_documents(
                self.gc, slide_id=self.slide_id, labeled=labeled)
            for doc in annotation_docs:
                _ = self.gc.post(
                    '/annotation?itemId=' + self.slide_id, json=doc)

        # Find size relative to WSI
        self.slide_info['F_tissue'] = self.slide_info[
            'sizeX'] / labeled.shape[1]

        return labeled



def convert_image_to_matrix(im):
    """Convert an image (MxNx3 array) to a column matrix of pixels
    (3x(M*N)).  It will pass through a 2D array unchanged.

    """
    if im.ndim == 2:
        return im

    return im.reshape((-1, im.shape[-1])).T



import numpy as np

from .condense import condense


def split(im_label, conn=8):
    """Re-labels objects that have multiple non-contiguous portions to create
    a new label image where each object is contiguous.

    Parameters
    ----------
    im_label : array_like
        A uint32 type label image generated by segmentation methods.
    conn : int
        Neighborhood connectivity to define contiguity. Valid values are 4 or
        8. Default value = 4.

    Notes
    -----
    Objects are assumed to have positive nonzero values.

    Returns
    -------
    Split : array_like
        A uint32 label where discontiguous objects are split and relabeled.

    See Also
    --------
    histomicstk.segmentation.label.area_open,
    histomicstk.segmentation.label.condense,
    histomicstk.segmentation.label.shuffle

    """
    import scipy.ndimage as ndi

    # copy input image
    Split = im_label.copy()

    # define kernel for neighborhood
    if conn == 8:
        Kernel = np.ones((3, 3), dtype=bool)
    elif conn == 4:
        Kernel = np.zeros((3, 3), dtype=bool)
        Kernel[1, :] = True
        Kernel[:, 1] = True
    else:
        msg = "Input 'conn' must be 4 or 8"
        raise ValueError(msg)

    # condense label image
    if np.unique(Split).size - 1 != Split.max():
        Split = condense(Split)

    # get locations of objects in initial image
    Locations = ndi.find_objects(Split)

    # initialize number of labeled objects
    Total = Split.max()

    # iterate through objects, replicating where needed
    for i in np.arange(1, len(Locations) + 1):

        # extract object from label image
        Template = Split[Locations[i - 1]]

        # label mask of object 'i'
        L, Count = ndi.label(Template == i, Kernel)

        # relabel if necessary
        if Count > 1:
            Template[L == 1] = i
            for i in np.arange(2, Count + 1):
                Template[L == i] = Total + 1
                Total += 1

    return Split



"""
Created on Mon Aug 12 18:33:48 2019.

@author: tageldim

"""

import cv2
import numpy as np
from shapely.geometry.polygon import Polygon

from histomicstk.utils.general_utils import Print_and_log


def get_contours_from_bin_mask(bin_mask):
    """Given a binary mask, get opencv contours.

    Parameters
    ----------
    bin_mask : nd array
        ground truth mask (m,n) - int32 with [0, 1] values.

    Returns
    -------
    dict
        a dictionary with the following keys:
        - contour group: the actual contour x,y coordinates.
        - hierarchy: contour hierarchy. This contains information about
        how contours relate to each other, in the form:
        [Next, Previous, First_Child, Parent,
        index_relative_to_contour_group]
        The last column is added for convenience and is not part of the
        original opencv output.
        - outer_contours: index of contours that do not have a parent, and
        are therefore the outermost most contours. These may have
        children (holes), however.
        See docs.opencv.org/3.1.0/d9/d8b/tutorial_py_contours_hierarchy.html
        for more information.

    """
    # Get contours using openCV. See this for modes:
    # https://docs.opencv.org/3.1.0/d9/d8b/tutorial_py_contours_hierarchy.html
    # we use the flag RETR_CCOMP so that we get boundary as well as holes
    # hierearchy output is: [Next, Previous, First_Child, Parent]
    ROI_cvuint8 = cv2.convertScaleAbs(bin_mask)
    cvout = cv2.findContours(
        ROI_cvuint8, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_SIMPLE)
    if len(cvout) < 3:
        contour_group, hierarchy = cvout[0], cvout[1]
    else:
        contour_group, hierarchy = cvout[1], cvout[2]
    hierarchy = hierarchy[0, ...]
    # outermost contours are those that don't have a parent
    # We'll append an index column to the rightmost end
    # to keep track of things better relative to contour_group, now it is:
    # [Next, Previous, First_Child, Parent, index_relative_to_contour_group]
    hierarchy = np.concatenate((hierarchy, np.arange(
        hierarchy.shape[0])[..., None]), axis=1)
    outer_contours = hierarchy[hierarchy[:, 3] == -1, :]
    conts = {
        'contour_group': contour_group,
        'hierarchy': hierarchy,
        'outer_contours': outer_contours,
    }
    return conts


def _add_contour_to_df(
        contours_df, mask_shape, conts, cidx, nest_info,
        pad_margin=0, MIN_SIZE=30, MAX_SIZE=None, monitorPrefix=''):
    """Add single contour to dataframe (Internal)."""
    # get coordinates for this contour. These are in x,y format.
    outer_cidx = conts['outer_contours'][cidx, 4]
    cont_outer = conts['contour_group'][outer_cidx][:, 0, :]
    if cont_outer.shape[0] <= 3:
        raise Exception('%s: TOO SIMPLE (%d coordinates) -- IGNORED' % (
            monitorPrefix, cont_outer.shape[0]))

    # Get index of first child (hole)
    inner_cidx = conts['outer_contours'][cidx, 2]
    has_holes = 0 + (inner_cidx > -1)

    # get nest location and size
    xmin, ymin = np.min(cont_outer, axis=0)
    nest_width, nest_height = np.max(
        cont_outer, 0) - np.min(cont_outer, 0)
    ymax = ymin + nest_height
    xmax = xmin + nest_width

    # ignore nests that are too small
    if (nest_height < MIN_SIZE) or (nest_width < MIN_SIZE):
        raise Exception('%s: TOO SMALL (%d x %d pixels) -- IGNORED' % (
            monitorPrefix, nest_height, nest_width))

    # ignore extremely large nests -- THESE MAY CAUSE SEGMENTATION FAULTS
    if MAX_SIZE is not None:
        if (nest_height > MAX_SIZE) or (nest_width > MAX_SIZE):
            raise Exception(
                '%s: EXTREMELY LARGE NEST (%d x %d pixels) -- IGNORED'
                % (monitorPrefix, nest_height, nest_width))

    # assign bounding box location
    ridx = contours_df.shape[0]
    contours_df.loc[ridx, 'group'] = nest_info['group']
    contours_df.loc[ridx, 'color'] = nest_info['color']
    contours_df.loc[ridx, 'ymin'] = ymin - pad_margin
    contours_df.loc[ridx, 'ymax'] = ymax - pad_margin
    contours_df.loc[ridx, 'xmin'] = xmin - pad_margin
    contours_df.loc[ridx, 'xmax'] = xmax - pad_margin

    # add other properties -- maybe useful later
    contours_df.loc[ridx, 'has_holes'] = has_holes
    contours_df.loc[ridx, 'touches_edge-top'] = 0 + (
        ymin - pad_margin - 2 < 0)
    contours_df.loc[ridx, 'touches_edge-left'] = 0 + (
        xmin - pad_margin - 2 < 0)
    contours_df.loc[ridx, 'touches_edge-bottom'] = 0 + (
        ymax + pad_margin + 2 > mask_shape[0])
    contours_df.loc[ridx, 'touches_edge-right'] = 0 + (
        xmax + pad_margin + 2 > mask_shape[1])

    # get x and y coordinates in HTK friendly format (string)
    cont_outer = conts['contour_group'][outer_cidx][:, 0, :].copy()
    contours_df.loc[ridx, 'coords_x'] = \
        ','.join([str(j - pad_margin) for j in list(cont_outer[:, 0])])
    contours_df.loc[ridx, 'coords_y'] = \
        ','.join([str(j - pad_margin) for j in list(cont_outer[:, 1])])

    return contours_df


def _get_contours_df(
        MASK, GTCodes_df, groups_to_get=None, MIN_SIZE=30, MAX_SIZE=None,
        verbose=False, monitorPrefix=''):
    """Parse ground truth mask and gets contours (Internal)."""
    from pandas import DataFrame

    cpr = Print_and_log(verbose=verbose)
    _print = cpr._print

    # pad with zeros to be able to detect edge contours later
    pad_margin = 50
    pad_value = 0
    while (GTCodes_df.GT_code == pad_value).any():
        pad_value += 1
    MASK = np.pad(MASK, pad_margin, 'constant', constant_values=pad_value)

    # Go through unique groups one by one -- each group (i.e. GTCode)
    # is extracted separately by binarizing the multi-class mask
    if groups_to_get is None:
        groups_to_get = list(GTCodes_df.index)
    else:
        groups_to_get = [
            GTCodes_df[GTCodes_df.group == group].head(1).index[0]
            if (GTCodes_df.group == group).any() else group
            for group in groups_to_get]
    contours_df = DataFrame()

    for nestgroup in groups_to_get:

        bin_mask = 0 + (MASK == GTCodes_df.loc[nestgroup, 'GT_code'])

        if bin_mask.sum() < MIN_SIZE * MIN_SIZE:
            _print('%s: %s: NO OBJECTS!!' % (monitorPrefix, nestgroup))
            continue

        _print('%s: %s: getting contours' % (monitorPrefix, nestgroup))
        conts = get_contours_from_bin_mask(bin_mask=bin_mask)
        n_tumor_nests = conts['outer_contours'].shape[0]

        # add nest contours
        _print('%s: %s: adding contours' % (monitorPrefix, nestgroup))
        for cidx in range(n_tumor_nests):
            try:
                nestcountStr = '%s: nest %s of %s' % (
                    monitorPrefix, cidx, n_tumor_nests)
                if cidx % 25 == 100:
                    _print(nestcountStr)

                contours_df = _add_contour_to_df(
                    contours_df, mask_shape=bin_mask.shape,
                    conts=conts, cidx=cidx,
                    nest_info=dict(GTCodes_df.loc[nestgroup, :]),
                    pad_margin=pad_margin, MIN_SIZE=MIN_SIZE,
                    MAX_SIZE=MAX_SIZE, monitorPrefix=nestcountStr)

            except Exception as e:
                _print(e)
                continue

    return contours_df


def _parse_annot_coords(annot, x_offset=0, y_offset=0):
    """Get x-, y- coordinates in a list format (Internal)."""
    coords_x = [int(j) + x_offset for j in annot['coords_x'].split(',')]
    coords_y = [int(j) + y_offset for j in annot['coords_y'].split(',')]
    coords = [(coords_x[i], coords_y[i]) for i in range(len(coords_x))]
    return coords


def _discard_nonenclosed_background_group(
        contours_df, background_group='mostly_stroma',
        verbose=False, monitorPrefix=''):
    """If a background group contour is NOT fully enclosed, discard it.

    This is a purely aesthetic method, makes sure that the background group
    contours (eg stroma) are discarded by default to avoid cluttering the
    field when posted to DSA for viewing online. The only exception is
    if they are enclosed within something else (eg tumor), in which case they
    are kept since they represent holes. This is related to
    https://github.com/DigitalSlideArchive/HistomicsTK/issues/675
    (Internal).

    """
    cpr = Print_and_log(verbose=verbose)
    _print = cpr._print

    # isolate background contours and non-background contours with holes
    background = contours_df.loc[
        contours_df.loc[:, 'group'] == background_group, :]
    contours_with_holes = contours_df.loc[
        contours_df.loc[:, 'group'] != background_group, :]
    contours_with_holes = contours_with_holes.loc[
        contours_with_holes.loc[:, 'has_holes'] == 1, :]

    def _append_polygon_if_valid(contDict, cid, polygon_list):
        try:
            polygon = Polygon(_parse_annot_coords(contDict))
            if polygon.is_valid:
                polygon_list.append(polygon)
        except Exception as e:
            _print('%s: contour %d: Shapely Error (below) -- IGNORED!' % (
                monitorPrefix, cid))
            _print(e)
        return polygon_list

    # to avoid redoing things, keep all non-background with holes in a list
    contour_polygons = []
    for cid, cont in contours_with_holes.iterrows():
        contour_polygons = _append_polygon_if_valid(
            dict(cont), cid=cid, polygon_list=contour_polygons)

    # iterate through stromal polygons and find if enclosed within something
    discard_cids = []
    for cid, cont in background.iterrows():
        bck_list = _append_polygon_if_valid(
            dict(cont), cid=cid, polygon_list=[])
        # only keep if enclosed with another contour
        discard = True
        if len(bck_list) > 0:
            for contour_polygon in contour_polygons:
                if contour_polygon.contains(bck_list[0]):
                    discard = False
        if discard:
            discard_cids.append(cid)

    # now drop unnecessary contours
    _print('%s: discarded %d contours' % (monitorPrefix, len(discard_cids)))
    contours_df.drop(discard_cids, axis=0, inplace=True)

    return contours_df


def get_contours_from_mask(
        MASK, GTCodes_df, groups_to_get=None, MIN_SIZE=30, MAX_SIZE=None,
        get_roi_contour=True, roi_group='roi',
        discard_nonenclosed_background=False, background_group='mostly_stroma',
        verbose=False, monitorPrefix=''):
    """Parse ground truth mask and gets contours for annotations.

    Parameters
    ----------
    MASK : nd array
        ground truth mask (m,n) where pixel values encode group membership.
    GTCodes_df : pandas Dataframe
        the ground truth codes and information dataframe.
        This is a dataframe that is indexed by the annotation group name and
        has the following columns.

        group: str
            group name of annotation, eg. mostly_tumor.
        GT_code: int
            desired ground truth code (in the mask). Pixels of this value
            belong to corresponding group (class).
        color: str
            rgb format. eg. rgb(255,0,0).
    groups_to_get : None
        if None (default) then all groups (ground truth labels) will be
        extracted. Otherwise pass a list of strings like ['mostly_tumor',].
    MIN_SIZE : int
        minimum bounding box size of contour
    MAX_SIZE : None
        if not None, int. Maximum bounding box size of contour. Sometimes
        very large contours cause segmentation faults that originate from
        opencv and are not caught by python, causing the python process
        to unexpectedly hault. If you would like to set a maximum size to
        defend against this, a suggested maximum would be 15000.
    get_roi_contour : bool
        whether to get contour for boundary of region of interest (ROI). This
        is most relevant when dealing with multiple ROIs per slide and with
        rotated rectangular or polygonal ROIs.
    roi_group : str
        name of roi group in the GT_Codes dataframe (eg roi)
    discard_nonenclosed_background : bool
        If a background group contour is NOT fully enclosed, discard it.
        This is a purely aesthetic method, makes sure that the background group
        contours (eg stroma) are discarded by default to avoid cluttering the
        field when posted to DSA for viewing online. The only exception is
        if they are enclosed within something else (eg tumor), in which case
        they are kept since they represent holes. This is related to
        https://github.com/DigitalSlideArchive/HistomicsTK/issues/675
        WARNING - This is a bit slower since the contours will have to be
        converted to shapely polygons. It is not noticeable for hundreds of
        contours, but you will notice the speed difference if you are parsing
        thousands of contours. Default, for this reason, is False.
    background_group : str
        name of background group in the GT_codes dataframe (eg mostly_stroma)
    verbose : bool
        Print progress to screen?
    monitorPrefix : str
        text to prepend to printed statements

    Returns
    -------
    pandas DataFrame
        contours extracted from input mask. The following columns are output.

        group : str
            annotation group (ground truth label).
        color : str
            annotation color if it were to be posted to DSA.
        is_roi : bool
            whether this annotation is a region of interest boundary
        ymin : int
            minimum y coordinate
        ymax : int
            maximum y coordinate
        xmin : int
            minimum x coordinate
        xmax : int
            maximum x coordinate
        has_holes : bool
            whether this contour has holes
        touches_edge-top : bool
            whether this contour touches top mask edge
        touches_edge-bottom : bool
            whether this contour touches bottom mask edge
        touches_edge-left : bool
            whether this contour touches left mask edge
        touches_edge-right : bool
            whether this contour touches right mask edge
        coords_x : str
            vertex x coordinates comma-separated values
        coords_y
            vertex y coordinated comma-separated values

    """
    from pandas import concat

    if MASK.sum() < 3:
        msg = 'Mask is empty!!'
        raise Exception(msg)
    cpr = Print_and_log(verbose=verbose)
    _print = cpr._print
    if groups_to_get is not None:
        _print("""WARNING!! Only specify groups_to_get is you do NOT mind
               having NO holes in polygons with holes that are occupied
               by a non-specified group. For example, let's say you
               specified that you only want to extract contours for
               tumor and stroma. If there is a large tumor polygon with two
               holes for stroma and blood vessel, the stroma hole will be
               accounted for, but not the blood vessel hole when you
               post these contours to DSA for viewing then pull them
               to be parse back into mask form. It's a subtle issue related
               to
               https://github.com/DigitalSlideArchive/HistomicsTK/issues/675
               and will eventually be accounted for once HistomicsTK
               has an official format to encode polygons with holes.""")

    cont_kwargs = {
        'GTCodes_df': GTCodes_df,
        'MIN_SIZE': MIN_SIZE,
        'MAX_SIZE': MAX_SIZE,
        'verbose': verbose,
    }

    # get contours df for non-roi contours
    contours_df = _get_contours_df(
        MASK=MASK, groups_to_get=groups_to_get,
        monitorPrefix='%s: %s' % (monitorPrefix, 'non-roi'),
        **cont_kwargs)

    # discard non-enclosed background (eg stroma) if needed
    if discard_nonenclosed_background:
        contours_df = _discard_nonenclosed_background_group(
            contours_df, background_group=background_group, verbose=verbose,
            monitorPrefix='%s: %s' % (monitorPrefix, 'discarding backgrnd'))

    # get contours df for roi boundary and concat
    if get_roi_contour:
        MASK_BIN = np.zeros(MASK.shape, dtype=np.uint8)
        MASK_BIN[MASK > 0] = GTCodes_df.loc[roi_group, 'GT_code']
        contours_df_roi = _get_contours_df(
            MASK=MASK_BIN, groups_to_get=[roi_group],
            monitorPrefix='%s: %s' % (monitorPrefix, roi_group),
            **cont_kwargs)
        contours_df = concat(
            (contours_df_roi, contours_df), axis=0, ignore_index=True)

    return contours_df


def get_single_annotation_document_from_contours(
        contours_df_slice, docname='default',
        F=1.0, X_OFFSET=0, Y_OFFSET=0, opacity=0.3,
        lineWidth=4.0, verbose=True, monitorPrefix=''):
    """Given dataframe of contours, get annotation document.

    This uses the large_image annotation schema to create an annotation
    document that maybe posted to DSA for viewing using something like:
    resp = gc.post("/annotation?itemId=" + slide_id, json=annotation_doc)
    The annotation schema can be found at:
    github.com/girder/large_image/blob/master/docs/annotations.md .

    Parameters
    ----------
    contours_df_slice : pandas DataFrame
        The following columns are of relevance and must be contained.

        group : str
            annotation group (ground truth label).
        color : str
            annotation color if it were to be posted to DSA.
        coords_x : str
            vertex x coordinates comma-separated values
        coords_y
            vertex y coordinated comma-separated values
    docname : str
        annotation document name
    F : float
        how much smaller is the mask where the contours come from is relative
        to the slide scan magnification. For example, if the mask is at 10x
        whereas the slide scan magnification is 20x, then F would be 2.0.
    X_OFFSET : int
        x offset to add to contours at BASE (SCAN) magnification
    Y_OFFSET : int
        y offset to add to contours at BASE (SCAN) magnification
    opacity : float
        opacity of annotation elements (in the range [0, 1])
    lineWidth : float
        width of boarders of annotation elements
    verbose : bool
        Print progress to screen?
    monitorPrefix : str
        text to prepend to printed statements

    Returns
    -------
    dict
        DSA-style annotation document ready to be post for viewing.

    """
    cpr = Print_and_log(verbose=verbose)
    _print = cpr._print

    def _get_fillColor(lineColor):
        fillColor = lineColor.replace('rgb', 'rgba')
        return fillColor[:fillColor.rfind(')')] + ',%.1f)' % opacity

    # Init annotation document in DSA style
    annotation_doc = {'name': docname, 'description': '', 'elements': []}

    # go through nests
    nno = 0
    nnests = contours_df_slice.shape[0]
    for _, nest in contours_df_slice.iterrows():

        nno += 1
        nestStr = '%s: contour %d of %s' % (monitorPrefix, nno, nnests)
        _print(nestStr)

        # Parse coordinates
        try:
            x_coords = F * np.int32(
                [int(j) for j in nest['coords_x'].split(',')]) + X_OFFSET
            y_coords = F * np.int32(
                [int(j) for j in nest['coords_y'].split(',')]) + Y_OFFSET
            zeros = np.zeros(x_coords.shape, dtype=np.int32)
            coords = np.concatenate(
                (x_coords[:, None], y_coords[:, None], zeros[:, None]),
                axis=1)
            coords = coords.tolist()
            coords.append(coords[0])
        except Exception as e:
            _print('%s: ERROR (below) - moving on!!!' % nestStr)
            _print(e)
            continue

        # assign to annotation style. See:
        # github.com/girder/large_image/blob/master/docs/annotations.md
        annotation_style = {
            'group': nest['group'],
            'type': 'polyline',
            'lineColor': nest['color'],
            'lineWidth': lineWidth,
            'closed': True,
            'points': coords,
            'label': {'value': nest['label']},
        }
        if opacity > 0:
            annotation_style['fillColor'] = _get_fillColor(nest['color'])

        # append to document
        annotation_doc['elements'].append(annotation_style)

    return annotation_doc


def get_annotation_documents_from_contours(
        contours_df, separate_docs_by_group=True, annots_per_doc=200,
        annprops=None, docnamePrefix='', verbose=True, monitorPrefix=''):
    """Given dataframe of contours, get list of annotation documents.

    This method parses a dataframe of contours to a list of dictionaries, each
    of which represents and large_image style annotation. This is a wrapper
    that extends the functionality of the method
    get_single_annotation_document_from_contours(), whose docstring should
    be referenced for implementation details and further explanation.

    Parameters
    ----------
    contours_df : pandas DataFrame
        WARNING - This is modified inside the function, so pass a copy.
        This dataframe includes data on contours extracted from input mask
        using get_contours_from_mask(). If you have contours using some other
        method, just make sure the dataframe follows the same schema as the
        output from get_contours_from_mask(). You may find a sample dataframe
        in the repo at
        ./tests/test_files/annotations_and_masks/sample_contours_df.tsv.
        The following columns are relevant for this method.

        group : str
            annotation group (ground truth label).
        color : str
            annotation color if it were to be posted to DSA.
        coords_x : str
            vertex x coordinates comma-separated values
        coords_y
            vertex y coordinated comma-separated values
    separate_docs_by_group : bool
        if set to True, you get one or more annotation documents (dicts)
        for each group (eg tumor) independently.
    annots_per_doc : int
        maximum number of annotation elements (polygons) per dict. The smaller
        this number, the more numerous the annotation documents, but the more
        seamless it is to post this data to the DSA server or to view using the
        HistomicsTK interface since you will be loading smaller chunks of data
        at a time.
    annprops : dict
        properties of annotation elements. Contains the following keys
        F, X_OFFSET, Y_OFFSET, opacity, lineWidth. Refer to
        get_single_annotation_document_from_contours() for details.
    docnamePrefix : str
        test to prepend to annotation document name
    verbose : bool
        Print progress to screen?
    monitorPrefix : str
        text to prepend to printed statements

    Returns
    -------
    list of dicts
        DSA-style annotation document.

    """
    if annprops is None:
        annprops = {
            'F': 1.0,
            'X_OFFSET': 0,
            'Y_OFFSET': 0,
            'opacity': 0,
            'lineWidth': 4.0,
        }
    if separate_docs_by_group:
        contours_df.loc[:, 'doc_group'] = contours_df.loc[:, 'group']
    else:
        contours_df.loc[:, 'doc_group'] = 'default'

    if 'label' not in contours_df.columns:
        contours_df.loc[:, 'label'] = contours_df.loc[:, 'group']

    # Each style goes to separate document(s) if sepate_docs_by_group
    annotation_docs = []
    for doc_group in set(contours_df.loc[:, 'doc_group']):

        # separate annotations with this group
        contours_df_slice = contours_df.loc[
            contours_df.loc[:, 'doc_group'] == doc_group, :]

        # Add every N annotations to a separate document
        if contours_df_slice.shape[0] > annots_per_doc:
            docbounds = list(range(
                0, contours_df_slice.shape[0], annots_per_doc))
            docbounds[-1] = contours_df_slice.shape[0]
        else:
            docbounds = [0, contours_df_slice.shape[0]]

        for docidx in range(len(docbounds) - 1):
            docStr = '%s: %s: doc %d of %d' % (
                monitorPrefix, doc_group, docidx + 1, len(docbounds) - 1)
            start = docbounds[docidx]
            end = docbounds[docidx + 1]
            annotation_doc = get_single_annotation_document_from_contours(
                contours_df_slice.iloc[start:end, :],
                docname='%s_%s-%d' % (docnamePrefix, doc_group, docidx),
                verbose=verbose, monitorPrefix=docStr, **annprops)
            if len(annotation_doc['elements']) > 0:
                annotation_docs.append(annotation_doc)

    return annotation_docs



"""
Created on Wed Sep 18 03:29:24 2019.

@author: mtageld
"""

import cv2
import numpy as np
from PIL import Image

from histomicstk.annotations_and_masks.annotation_and_mask_utils import \
    get_image_from_htk_response
from histomicstk.annotations_and_masks.masks_to_annotations_handler import (
    get_annotation_documents_from_contours, get_contours_from_mask)
from histomicstk.preprocessing.color_deconvolution.color_deconvolution import \
    color_deconvolution_routine

Image.MAX_IMAGE_PIXELS = None


def get_slide_thumbnail(gc, slide_id):
    """Get slide thumbnail using girder client.

    Parameters
    ----------
    gc : object
        girder client to use
    slide_id : str
        girder ID of slide

    Returns
    -------
    np array
        RGB slide thumbnail at lowest level

    """
    getStr = '/item/%s/tiles/thumbnail' % (slide_id)
    resp = gc.get(getStr, jsonResp=False)
    return get_image_from_htk_response(resp)


def _deconv_color(im, **kwargs):
    """Wrap around color_deconvolution_routine (compatibility)."""
    Stains, _, _ = color_deconvolution_routine(im, **kwargs)
    return Stains, 0


def get_tissue_mask(
        thumbnail_im,
        deconvolve_first=False, stain_unmixing_routine_kwargs=None,
        n_thresholding_steps=1, sigma=0., min_size=500):
    """Get binary tissue mask from slide thumbnail.

    Parameters
    ----------
    thumbnail_im : np array
        (m, n, 3) nd array of thumbnail RGB image
        or (m, n) nd array of thumbnail grayscale image
    deconvolve_first : bool
        use hematoxylin channel to find cellular areas?
        This will make things ever-so-slightly slower but is better in
        getting rid of sharpie marker (if it's green, for example).
        Sometimes things work better without it, though.
    stain_matrix_method : str
        see deconv_color method in seed_utils
    n_thresholding_steps : int
        number of gaussian smoothing steps
    sigma : float
        sigma of gaussian filter
    min_size : int
        minimum size (in pixels) of contiguous tissue regions to keep

    Returns
    -------
    np int32 array
        each unique value represents a unique tissue region
    np bool array
        largest contiguous tissue region.

    """
    from scipy import ndimage
    from skimage.filters import gaussian, threshold_otsu

    stain_unmixing_routine_kwargs = (
        {} if stain_unmixing_routine_kwargs is None else stain_unmixing_routine_kwargs)

    if deconvolve_first and (len(thumbnail_im.shape) == 3):
        # deconvolvve to ge hematoxylin channel (cellular areas)
        # hematoxylin channel return shows MINIMA so we invert
        stain_unmixing_routine_kwargs['stains'] = ['hematoxylin', 'eosin']
        Stains, _, _ = color_deconvolution_routine(
            thumbnail_im, **stain_unmixing_routine_kwargs)
        thumbnail = 255 - Stains[..., 0]

    elif len(thumbnail_im.shape) == 3:
        # grayscale thumbnail (inverted)
        thumbnail = 255 - cv2.cvtColor(thumbnail_im, cv2.COLOR_BGR2GRAY)

    else:
        thumbnail = thumbnail_im

    for _ in range(n_thresholding_steps):

        # gaussian smoothing of grayscale thumbnail
        if sigma > 0.0:
            thumbnail = gaussian(
                thumbnail, sigma=sigma,
                out=None, mode='nearest', preserve_range=True)

        # get threshold to keep analysis region
        try:
            thresh = threshold_otsu(thumbnail[thumbnail > 0])
        except ValueError:  # all values are zero
            thresh = 0

        # replace pixels outside analysis region with upper quantile pixels
        thumbnail[thumbnail < thresh] = 0

    # convert to binary
    mask = 0 + (thumbnail > 0)

    # find connected components
    labeled, _ = ndimage.label(mask)

    # only keep
    unique, counts = np.unique(labeled[labeled > 0], return_counts=True)
    discard = np.isin(labeled, unique[counts < min_size])
    discard = discard.reshape(labeled.shape)
    labeled[discard] = 0

    # largest tissue region
    mask = labeled == unique[np.argmax(counts)]

    return labeled, mask


def get_tissue_boundary_annotation_documents(
        gc, slide_id, labeled,
        color='rgb(0,0,0)', group='tissue', annprops=None):
    """Get annotation documents of tissue boundaries to visualize on DSA.

    Parameters
    ----------
    gc : object
        girder client to use
    slide_id : str
        girder ID of slide
    labeled : np array
        mask of tissue regions using slide thumbnail. This could either be
        a binary mask or a mask where each unique value corresponds to one
        tissue region. It will be binalized anyways. This can be obtained
        using get_tissue_mask().
    color : str
        color to assign to boundaries. format like rgb(0,0,0)
    group : str
        label for annotations
    annpops : dict
        properties of annotation elements. Contains the following keys
        F, X_OFFSET, Y_OFFSET, opacity, lineWidth. Refer to
        get_single_annotation_document_from_contours() for details.

    Returns
    -------
    list of dicts
        each dict is an annotation document that you can post to DSA

    """
    from pandas import DataFrame

    # Get annotations properties
    if annprops is None:
        slide_info = gc.get('item/%s/tiles' % slide_id)
        annprops = {
            'F': slide_info['sizeX'] / labeled.shape[1],  # relative to base
            'X_OFFSET': 0,
            'Y_OFFSET': 0,
            'opacity': 0,
            'lineWidth': 4.0,
        }

    # Define GTCodes dataframe
    GTCodes_df = DataFrame(columns=['group', 'GT_code', 'color'])
    GTCodes_df.loc['tissue', 'group'] = group
    GTCodes_df.loc['tissue', 'GT_code'] = 1
    GTCodes_df.loc['tissue', 'color'] = color

    # get annotation docs
    contours_tissue = get_contours_from_mask(
        MASK=0 + (labeled > 0), GTCodes_df=GTCodes_df,
        get_roi_contour=False, MIN_SIZE=0, MAX_SIZE=None, verbose=False,
        monitorPrefix='tissue: getting contours')
    annotation_docs = get_annotation_documents_from_contours(
        contours_tissue.copy(), docnamePrefix='test', annprops=annprops,
        verbose=False, monitorPrefix='tissue : annotation docs')

    return annotation_docs


def threshold_multichannel(
        im, thresholds, channels=None,
        just_threshold=False, get_tissue_mask_kwargs=None):
    """Threshold a multi-channel image (eg. HSI image) to get tissue.

    The relies on the fact that oftentimes some slide elements (eg blood
    or whitespace) have a characteristic hue/saturation/intensity. This
    thresholds along each HSI channel, then optionally uses the
    get_tissue_mask() method (gaussian smoothing, otsu thresholding,
    connected components) to get each contiguous tissue piece.

    Parameters
    ----------
    im : np array
        (m, n, 3) array of Hue, Saturation, Intensity (in this order)
    thresholds : dict
        Each entry is a dict containing the keys min and max
    channels : list
        names of channels, in order (eg. hue, saturation, intensity)
    just_threshold : bool
        if False, get_tissue_mask() is used to smooth result and get regions.
    get_tissue_mask_kwargs : dict
        key-value pairs of parameters to pass to get_tissue_mask()

    Returns
    -------
    np int32 array
        if not just_threshold, unique values represent unique tissue regions
    np bool array
        if not just_threshold, largest contiguous tissue region.

    """
    channels = ['hue', 'saturation', 'intensity'] if channels is None else channels

    if get_tissue_mask_kwargs is None:
        get_tissue_mask_kwargs = {
            'n_thresholding_steps': 1,
            'sigma': 5.0,
            'min_size': 10,
        }

    # threshold each channel
    mask = np.ones(im.shape[:2])
    for ax, ch in enumerate(channels):

        channel = im[..., ax].copy()

        mask[channel < thresholds[ch]['min']] = 0
        mask[channel >= thresholds[ch]['max']] = 0

    # smoothing, otsu thresholding then connected components
    if just_threshold or (np.unique(mask).shape[0] < 1):
        labeled = mask
    else:
        get_tissue_mask_kwargs['deconvolve_first'] = False
        labeled, mask = get_tissue_mask(mask, **get_tissue_mask_kwargs)

    return labeled, mask


def _get_largest_regions(labeled, top_n=10):

    labeled_im = labeled.copy()

    unique, counts = np.unique(labeled_im[labeled_im > 0], return_counts=True)

    keep = unique[np.argsort(counts)[-top_n:]]

    mask = np.zeros(labeled_im.shape)
    keep_pixels = np.isin(labeled_im, keep)
    keep_pixels = keep_pixels.reshape(labeled_im.shape)
    mask[keep_pixels] = 1
    labeled_im[mask == 0] = 0

    return labeled_im



import numpy as np

from histomicstk.utils import gradient_diffusion


def gvf_tracking(I, Mask, K=1000, Diffusions=10, Mu=5, Lambda=5, Iterations=10,
                 dT=0.05):
    """
    Performs gradient-field tracking to segment smoothed images of cell nuclei.

    Takes as input a smoothed intensity or Laplacian-of-Gaussian filtered image
    and a foreground mask, and groups pixels by tracking them to mutual
    gradient sinks. Typically requires merging of sinks (seeds) as a post
    processing steps.

    Parameters
    ----------
    I : array_like
        Smoothed intensity or log-filtered response where nuclei regions have
        larger intensity values than background.
    Mask : array_like
        Binary mask where foreground objects have value 1, and background
        objects have value 0. Used to restrict influence of background vectors
        on diffusion process and to reduce tracking computations.
    K : float
        Number of steps to check for tracking cycle. Default value = 1000.
    Mu : float
        Weight parameter from Navier-Stokes diffusion - weights divergence and
        Laplacian terms. Default value = 5.
    Lambda : float
        Weight parameter from Navier-Stokes diffusion - used to weight
        divergence. Default value = 5.
    Iterations : float
        Number of time-steps to use in Navier-Stokes diffusion. Default value =
        10.
    dT : float
        Timestep to be used in Navier-Stokes diffusion. Default value = 0.05.

    Returns
    -------
    Segmentation : array_like
        Label image where positive values correspond to foreground pixels that
        share mutual sinks.
    Sinks : array_like
        N x 2 array containing the (x,y) locations of the tracking sinks. Each
        row is an (x,y) pair - in that order.

    See Also
    --------
    histomicstk.utils.gradient_diffusion,
    histomicstk.segmentation.label.shuffle

    References
    ----------
    .. [#] G. Li et al "3D cell nuclei segmentation based on gradient flow
       tracking" in BMC Cell Biology,vol.40,no.8, 2007.

    """
    # get image shape
    M = I.shape[0]
    N = I.shape[1]

    # calculate gradient
    dy, dx = np.gradient(I)

    # diffusion iterations
    if Diffusions > 0:
        dx, dy = gradient_diffusion(dx, dy, Mask, Mu, Lambda, Diffusions,
                                    dT)

    # normalize to unit magnitude
    Mag = ((dx**2 + dy**2)**0.5 + np.finfo(float).eps)
    dy = dy / Mag
    dx = dx / Mag

    # define mask to track pixels that are mapped to a sink
    Mapped = np.zeros(I.shape)

    # define label image
    Segmentation = np.zeros(I.shape)

    # initialize lists of sinks
    Sinks = []

    # define coordinates for foreground pixels (Mask == 1)
    i, j = np.nonzero(Mask)

    # track pixels
    for _index, (x, y) in enumerate(zip(j, i)):

        # initialize angle, trajectory length, novel flag, and allocation count
        phi = 0
        points = 1
        novel = 1
        alloc = 1

        # initialize trajectory
        Trajectory = np.zeros((K, 2))
        Trajectory[0, 0] = x
        Trajectory[0, 1] = y

        # track while angle defined by successive steps is < np.pi / 2
        while phi < np.pi / 2:

            # calculate step
            xStep = round_float(dx[Trajectory[points - 1, 1],
                                Trajectory[points - 1, 0]])
            yStep = round_float(dy[Trajectory[points - 1, 1],
                                Trajectory[points - 1, 0]])

            # check image edge
            if ((Trajectory[points - 1, 0] + xStep < 0) or
                    (Trajectory[points - 1, 0] + xStep > N - 1) or
                    (Trajectory[points - 1, 1] + yStep < 0) or
                    (Trajectory[points - 1, 1] + yStep > M - 1)):
                break

            # add new point to trajectory list
            if points < K:  # buffer is not overrun
                Trajectory[points, 0] = Trajectory[points - 1, 0] + xStep
                Trajectory[points, 1] = Trajectory[points - 1, 1] + yStep

            else:  # buffer overrun

                # check for cycle
                cycle = detect_cycle(Trajectory, points)

                if cycle == points:  # no cycle, simple overflow. grow buffer.

                    # copy and reallocate
                    temp = Trajectory
                    Trajectory = np.zeros((K * alloc, 2))
                    Trajectory[K * (alloc - 1):K * alloc, ] = temp
                    alloc += 1

                    # add new point
                    Trajectory[points, 0] = Trajectory[points - 1, 0] + xStep
                    Trajectory[points, 1] = Trajectory[points - 1, 1] + yStep

                else:  # overflow due to cycle, terminate tracking
                    points = cycle

            # check mapping
            if Mapped[Trajectory[points, 1], Trajectory[points, 0]] == 1:
                novel = 0
                phi = np.pi
            elif Mask[Trajectory[points, 1], Trajectory[points, 0]] == 0:
                phi = np.pi
            else:
                phi = np.arccos(dy[Trajectory[points - 1, 1],
                                   Trajectory[points - 1, 0]] *
                                dy[Trajectory[points, 1],
                                   Trajectory[points, 0]] +
                                dx[Trajectory[points - 1, 1],
                                   Trajectory[points - 1, 0]] *
                                dx[Trajectory[points, 1],
                                   Trajectory[points, 0]])

            # increment trajectory length counter
            points += 1

        # determine if sink is novel
        if novel == 1:

            # record sinks
            Sinks.append(Trajectory[points - 1, ])

            # add trajectory to label image with new sink value, add mapping
            for j in range(points):
                Segmentation[Trajectory[j, 1], Trajectory[j, 0]] = len(Sinks)
                Mapped[Trajectory[j, 1], Trajectory[j, 0]] = 1

        else:

            # add trajectory to label image with sink value of final point
            for j in range(points):
                Segmentation[Trajectory[j, 1], Trajectory[j, 0]] = \
                    Segmentation[Trajectory[points - 1, 1],
                                 Trajectory[points - 1, 0]]

    # convert Sinks to numpy array
    Sinks = np.asarray(Sinks)

    return Segmentation, Sinks


def merge_sinks(Label, Sinks, Radius=5):
    """
    Merges attraction basins obtained from gradient flow tracking using
    sink locations.

    Parameters
    ----------
    Segmentation : array_like
        Label image where positive values correspond to foreground pixels that
        share mutual sinks.
    Sinks : array_like
        N x 2 array containing the (x,y) locations of the tracking sinks. Each
        row is an (x,y) pair - in that order.
    Radius : float
        Radius used to merge sinks. Sinks closer than this radius to one
        another will have their regions of attraction merged.
        Default value = 5.

    Returns
    -------
    Merged : array_like
        Label image where attraction regions are merged.

    """
    import skimage.morphology as mp
    from skimage import measure as ms

    # build seed image
    SeedImage = np.zeros(Label.shape)
    for i in range(Sinks.shape[0]):
        SeedImage[Sinks[i, 1], Sinks[i, 0]] = i + 1

    # dilate sink image
    Dilated = mp.binary_dilation(SeedImage, mp.disk(Radius))

    # generate new labels for merged seeds, define memberships
    Labels = ms.label(Dilated)
    New = Labels[Sinks[:, 1].astype(int), Sinks[:, 0].astype(int)]

    # get unique list of seed clusters
    Unique = np.arange(1, New.max() + 1)

    # generate new seed list
    Merged = np.zeros(Label.shape)

    # get pixel list for each sink object
    Props = ms.regionprops(Label.astype(int))

    # fill in new values
    for i in Unique:
        Indices = np.nonzero(New == i)[0]
        for j in Indices:
            Coords = Props[j].coords
            Merged[Coords[:, 0], Coords[:, 1]] = i

    return Merged


def detect_cycle(Trajectory, points):

    # initialize trajectory length
    length = 0

    # identify trajectory bounding box
    xMin = np.min(Trajectory[0:points, 0])
    xMax = np.max(Trajectory[0:points, 0])
    xRange = xMax - xMin + 1
    yMin = np.min(Trajectory[0:points, 1])
    yMax = np.max(Trajectory[0:points, 1])
    yRange = yMax - yMin + 1

    # fill in trajectory map
    Map = np.zeros((yRange, xRange))
    for i in range(points):
        if Map[Trajectory[i, 1] - yMin, Trajectory[i, 0] - xMin] == 1:
            break
        else:
            Map[Trajectory[i, 1] - yMin, Trajectory[i, 0] - xMin] = 1
        length += 1

    return length


def round_float(x):
    if x >= 0.0:
        t = np.ceil(x)
        if t - x > 0.5:
            t -= 1.0
        return t
    else:
        t = np.ceil(-x)
        if t + x > 0.5:
            t -= 1.0
        return -t



import json
import math
import os
from pathlib import Path

import large_image
import numpy
import scipy
import skimage

import histomicstk
from histomicstk.cli import utils
from histomicstk.cli.utils import CLIArgumentParser


def createSuperPixels(opts):  # noqa
    """
    Besides the inputs described by the specification, this also can take
    opts.callback, which is a function that takes (step_name: str,
    step_count: int, step_total: int).
    """
    import pyvips

    averageSize = opts.superpixelSize ** 2
    overlap = opts.superpixelSize * 4 * 2 if opts.overlap else 0
    tileSize = opts.tileSize + overlap

    print('>> Reading input image')
    print(opts.inputImageFile)

    ts = large_image.open(opts.inputImageFile)
    meta = ts.getMetadata()
    found = 0
    strips = []
    bboxes = []
    bboxesUser = []
    tiparams = {}
    tiparams = utils.get_region_dict(opts.roi, None, ts)

    scale = 1
    if opts.magnification:
        tiparams['scale'] = {'magnification': opts.magnification}

    print('>> Generating superpixels')
    if opts.slic_zero:
        print('>> Using SLIC Zero for segmentation')
    for tile in ts.tileIterator(
        format=large_image.constants.TILE_FORMAT_NUMPY,
        tile_size=dict(width=tileSize, height=tileSize),
        tile_overlap=dict(x=overlap, y=overlap),
        **tiparams,
    ):
        if hasattr(opts, 'callback'):
            opts.callback('tiles', tile['tile_position']['position'],
                          tile['iterator_range']['position'])
        print('%d/%d (%d x %d) - %d' % (
            tile['tile_position']['position'], tile['iterator_range']['position'],
            tile['width'], tile['height'],
            found))
        if meta['magnification'] and tile['magnification']:
            scale = meta['magnification'] / tile['magnification']
        x0 = tiparams.get('region', {}).get('left', 0)
        y0 = tiparams.get('region', {}).get('top', 0)
        # tx0 = tile['x'] - x0
        # ty0 = tile['y'] - y0
        tx0 = int((tile['gx'] - x0) / scale)
        ty0 = int((tile['gy'] - y0) / scale)
        img = tile['tile']
        n_pixels = tile['width'] * tile['height']
        mask = None
        if overlap:
            mask = numpy.ones(img.shape[:2])
            for y, simg in strips:
                if (y < ty0 + tile['height'] and y + simg.height > ty0 and simg.width > tx0):
                    suby = max(0, y - ty0)
                    subimg = simg.crop(
                        tx0,
                        max(0, ty0 - y),
                        min(tile['width'], simg.width),
                        min(tile['height'], simg.height - max(0, ty0 - y)))
                    # Our mask is true when a pixel has not been set
                    submask = numpy.ndarray(
                        buffer=subimg[3].write_to_memory(),
                        dtype=numpy.uint8,
                        shape=[subimg.height, subimg.width]) == 0
                    mask[suby:suby + submask.shape[0], :submask.shape[1]] *= submask
            n_pixels = numpy.count_nonzero(mask)
        n_segments = math.ceil(n_pixels / averageSize)
        segments = skimage.segmentation.slic(
            img,
            n_segments=n_segments,
            slic_zero=bool(opts.slic_zero),
            compactness=opts.compactness,
            sigma=opts.sigma,
            start_label=0,
            enforce_connectivity=True,
            mask=mask,
        )
        # We now have an array that is the same size as the image
        maxValue = numpy.max(segments) + 1
        if overlap:
            # Keep any segment that is at all in the non-overlap region
            core = segments[
                :tile['height'] - tile['tile_overlap']['bottom'],
                :tile['width'] - tile['tile_overlap']['right']]
            coremask = mask[
                :tile['height'] - tile['tile_overlap']['bottom'],
                :tile['width'] - tile['tile_overlap']['right']]
            core[numpy.where(coremask != 1)] = -1
            usedIndices = numpy.unique(core)
            usedIndices = numpy.delete(usedIndices, numpy.where(usedIndices < 0))
            usedLut = [-1] * maxValue
            for idx, used in enumerate(usedIndices):
                if used >= 0:
                    usedLut[used] = idx
            usedLut = numpy.array(usedLut, dtype=int)
            print('reduced from %d to %d' % (maxValue, len(usedIndices)))
            maxValue = len(usedIndices)
            segments = usedLut[segments]
            mask *= (segments != -1)
        if str(opts.bounding).lower() not in {'', 'none'}:
            regions = skimage.measure.regionprops(1 + segments)
            for _pidx, props in enumerate(regions):
                by0, bx0, by1, bx1 = props.bbox
                bboxes.append((
                    ((bx0 + bx1) / 2 + tx0) * scale + x0,
                    ((by0 + by1) / 2 + ty0) * scale + y0,
                    (bx1 - bx0) * scale,
                    (by1 - by0) * scale))
                bboxesUser.extend([
                    (bx0 + tx0) * scale + x0,
                    (by0 + ty0) * scale + y0,
                    (bx1 + tx0) * scale + x0,
                    (by1 + ty0) * scale + y0,
                ])
        if opts.boundaries:
            segments *= 2
            maxValue *= 2
            edges = (scipy.ndimage.sobel(segments, axis=0) != 0) | (
                scipy.ndimage.sobel(segments, axis=1) != 0)
            edges[0, :] = True
            edges[-1, :] = True
            edges[:, 0] = True
            edges[:, -1] = True
            segments += edges
        segments += found
        found += int(maxValue)
        if mask is None:
            data = numpy.dstack((
                (segments % 256).astype(int),
                (segments / 256).astype(int) % 256,
                (segments / 65536).astype(int) % 256)).astype('B')
        else:
            data = numpy.dstack((
                (segments % 256).astype(int),
                (segments / 256).astype(int) % 256,
                (segments / 65536).astype(int) % 256,
                mask * 255)).astype('B')
        # For overlay, suppose we make any value whose centroid is in the
        # overlap region transparent.  Then, use vips to overlap the
        # images rather than inserting them.
        vimg = pyvips.Image.new_from_memory(
            numpy.ascontiguousarray(data).data,
            data.shape[1], data.shape[0], data.shape[2],
            large_image.constants.dtypeToGValue[data.dtype.char])
        vimg = vimg.copy(interpretation=pyvips.Interpretation.RGB)
        vimgTemp = pyvips.Image.new_temp_file('%s.v')
        vimg.write(vimgTemp)
        vimg = vimgTemp
        x = tx0
        ty = tile['tile_position']['region_y']
        while len(strips) <= ty:
            strips.append(None)
        if strips[ty] is None:
            strip = pyvips.Image.black(
                tiparams.get('region', {}).get('width', meta['sizeX']),
                vimg.height, bands=vimg.bands)
            strip = strip.copy(interpretation=pyvips.Interpretation.RGB)
            strips[ty] = [ty0, strip]
        strips[ty][1] = strips[ty][1].composite([vimg], pyvips.BlendMode.OVER, x=int(x), y=0)
        if hasattr(opts, 'callback'):
            opts.callback('tiles', tile['tile_position']['position'] + 1,
                          tile['iterator_range']['position'])
    if hasattr(opts, 'callback'):
        opts.callback('file', 0, 2 if opts.outputAnnotationFile else 1)
    print('>> Found %d superpixels' % found)
    if found > 256 ** 3:
        print('Too many superpixels')
    img = pyvips.Image.black(
        tiparams.get('region', {}).get('width', meta['sizeX']) / scale,
        tiparams.get('region', {}).get('height', meta['sizeY']) / scale,
        bands=strips[0][1].bands)
    img = img.copy(interpretation=pyvips.Interpretation.RGB)
    for stripidx in range(len(strips)):
        img = img.composite(
            [strips[stripidx][1]], pyvips.BlendMode.OVER, x=0, y=int(strips[stripidx][0]))
    # Discard alpha band, if any.
    img = img[:3]
    # Add program run parameters to the image description and list the
    # superpixel count
    img.set_type(
        pyvips.GValue.gstr_type, 'image-description',
        json.dumps(dict(
            {k: v for k, v in vars(opts).items() if k != 'callback'}, indexCount=found)))
    img.write_to_file(
        opts.outputImageFile, tile=True, tile_width=256, tile_height=256, pyramid=True,
        region_shrink=pyvips.RegionShrink.NEAREST,
        # We'd prefer max, but to do so we need to compute max of the
        # superpixel, not the faux-color it is mapped to.
        # region_shrink=pyvips.RegionShrink.MAX,
        bigtiff=True, compression='lzw', predictor='horizontal')

    if hasattr(opts, 'callback'):
        opts.callback('file', 1, 2 if opts.outputAnnotationFile else 1)
    if opts.outputAnnotationFile:
        print('>> Generating annotation file')
        categories = [
            {
                'label': opts.default_category_label,
                'fillColor': opts.default_fillColor,
                'strokeColor': opts.default_strokeColor,
            },
        ]
        annotation_name = os.path.splitext(os.path.basename(opts.outputAnnotationFile))[0]
        region_dict = utils.get_region_dict(opts.roi, None, ts)
        annotation = {
            'name': annotation_name,
            'elements': [{
                'type': 'pixelmap',
                'girderId': 'outputImageFile',
                'transform': {
                    'xoffset': region_dict.get('region', {}).get('left', 0) / scale,
                    'yoffset': region_dict.get('region', {}).get('top', 0) / scale,
                    'matrix': [[scale, 0], [0, scale]],
                },
                'values': [0] * (found // (2 if opts.boundaries else 1)),
                'categories': categories,
                'boundaries': opts.boundaries,
            }],
            'attributes': {
                'params': {k: v for k, v in vars(opts).items() if not callable(v)},
                'cli': Path(__file__).stem,
                'version': histomicstk.__version__,
            },
        }
        if len(bboxes) and str(opts.bounding).lower() != 'separate':
            annotation['elements'][0]['user'] = {'bbox': bboxesUser}
        if len(bboxes) and str(opts.bounding).lower() != 'internal':
            bboxannotation = {
                'name': '%s bounding boxes' % os.path.splitext(
                    os.path.basename(opts.outputAnnotationFile))[0],
                'elements': [{
                    'type': 'rectangle',
                    'center': [bcx, bcy, 0],
                    'width': bw,
                    'height': bh,
                    'rotation': 0,
                    'label': {'value': 'Region %d' % bidx},
                    'fillColor': 'rgba(0,0,0,0)',
                    'lineColor': opts.default_strokeColor,
                } for bidx, (bcx, bcy, bw, bh) in enumerate(bboxes)],
                'attributes': {
                    'params': {k: v for k, v in vars(opts).items() if not callable(v)},
                    'cli': Path(__file__).stem,
                    'version': histomicstk.__version__,
                },
            }
            annotation = [annotation, bboxannotation]
        with open(opts.outputAnnotationFile, 'w') as annotation_file:
            try:
                json.dump(annotation, annotation_file, separators=(',', ':'), sort_keys=False)
            except Exception:
                print('Failed to serialize annotation')
                print(repr(annotation))
                raise
        if hasattr(opts, 'callback'):
            opts.callback('file', 2, 2)


if __name__ == '__main__':
    createSuperPixels(CLIArgumentParser().parse_args())



import numpy as np


def rgb_to_sda(im_rgb, I_0, allow_negatives=False):
    """Transform input RGB image or matrix `im_rgb` into SDA (stain
    darkness) space for color deconvolution.

    Parameters
    ----------
    im_rgb : array_like
        Image (MxNx3) or matrix (3xN) of pixels

    I_0 : float or array_like
        Background intensity, either per-channel or for all channels

    allow_negatives : bool
        If False, would-be negative values in the output are clipped to 0

    Returns
    -------
    im_sda : array_like
        Shaped like `im_rgb`, with output values 0..255 where `im_rgb` >= 1

    Note
    ----
    For compatibility purposes, passing I_0=None invokes the behavior of
    rgb_to_od.

    See Also
    --------
    histomicstk.preprocessing.color_conversion.sda_to_rgb,
    histomicstk.preprocessing.color_conversion.rgb_to_od,
    histomicstk.preprocessing.color_deconvolution.color_deconvolution,
    histomicstk.preprocessing.color_deconvolution.color_convolution

    """
    is_matrix = im_rgb.ndim == 2
    if is_matrix:
        im_rgb = im_rgb.T

    if I_0 is None:  # rgb_to_od compatibility
        im_rgb = im_rgb.astype(float) + 1
        I_0 = 256

    im_rgb = np.maximum(im_rgb, 1e-10)

    im_sda = -np.log(im_rgb / (1. * I_0)) * 255 / np.log(I_0)
    if not allow_negatives:
        im_sda = np.maximum(im_sda, 0)
    return im_sda.T if is_matrix else im_sda



"""
Created on Tue Oct  1 01:38:16 2019.

@author: mtageld
"""
from histomicstk.annotations_and_masks.annotation_and_mask_utils import \
    delete_annotations_in_slide


def cellularity_detection_workflow(
        gc, cdo, slide_id, monitorPrefix='',
        destination_folder_id=None, keep_existing_annotations=False):
    """Run cellularity detection for single slide.

    The cellularity detection algorithm can either be
    Cellularity_detector_superpixels or Cellularity_detector_thresholding.

    Arguments:
    ---------
    gc : object
        girder client object
    cdo : object
        Cellularity_detector object instance. Can either be
        Cellularity_detector_superpixels() or
        Cellularity_detector_thresholding(). The thresholding-based workflow
        seems to be more robust, despite being simpler.
    slide_id : str
        girder id of slide on which workflow is done
    monitorPrefix : str
        this will set the cds monitorPrefix attribute
    destination_folder_id : str or None
        if not None, copy slide to this girder folder and post results
        there instead of original slide.
    keep_existing_annotations : bool
        keep existing annotations in slide when posting results?

    """
    cdo.monitorPrefix = monitorPrefix

    # copy slide to target folder, otherwise work in-place
    if destination_folder_id is not None:
        cdo._print1('%s: copying slide to destination folder' % monitorPrefix)
        resp = gc.post(
            '/item/%s/copy?folderId=%s&copyAnnotations=%s' %
            (slide_id, destination_folder_id, keep_existing_annotations))
        slide_id = resp['_id']

    elif not keep_existing_annotations:
        cdo._print1('%s: deleting existing annotations' % monitorPrefix)
        delete_annotations_in_slide(gc, slide_id)

    # run cds for this slide
    cdo.slide_id = slide_id
    cdo.run()

    return slide_id



import large_image
import numpy as np


def compute_tile_foreground_fraction(slide_path, im_fgnd_mask_lres,
                                     fgnd_seg_scale, it_kwargs,
                                     tile_position=None, style=None):
    """
    Computes the fraction of foreground of a single tile or
    all tiles in a whole slide image given the binary foreground
    mask computed from a low resolution version of the slide.

    Parameters
    ----------
    slide_path : str
        path to an image or slide
    im_fgnd_mask_lres : array_like
        A binary foreground mask computed at a low-resolution
    fgnd_seg_scale : double
        The scale/magnification at which the foreground mask `im_fgnd_mask_lres`
        was computed
    it_kwargs : dict
        A dictionary of any key:value parameters (e.g. defining the scale,
         tile_size, region etc) in addition to tile_position that need to be
         passed to `large_image.TileSource.getSingleTile` to get the tile.
    tile_position : int or None
        A linear 0-based index of a tile for which the foreground needs to be
        computed. If set to None, the foreground fraction of all tiles will be
        computed.

    Returns
    -------
    tile_fgnd_frac : double or array_like
        A value between 0 and 1 indicating the fraction of foreground pixels
        present in the tile indicated by `tile_position`. If `tile_position`
        is set to None, then a 1D array containing the foreground fraction of
        all tiles will be returned.

    """
    import dask
    import dask.distributed

    if tile_position is None:

        # get slide tile source
        ts = large_image.getTileSource(slide_path, style=style)

        num_tiles = ts.getSingleTile(**it_kwargs)['iterator_range']['position']

        # broadcasting fgnd mask to all dask workers
        try:
            c = dask.distributed.get_client()

            [im_fgnd_mask_lres] = c.scatter([im_fgnd_mask_lres],
                                            broadcast=True)
        except ValueError:
            pass

        # compute tile foreground fraction in parallel
        tile_fgnd_frac = []

        for tile_position in range(num_tiles):

            tile_fgnd_frac.append(
                dask.delayed(_compute_tile_foreground_fraction_single)(
                    slide_path, im_fgnd_mask_lres, fgnd_seg_scale,
                    it_kwargs, tile_position,
                ))

        tile_fgnd_frac = dask.delayed(tile_fgnd_frac).compute()

        tile_fgnd_frac = np.array(tile_fgnd_frac)

    elif np.isscalar(tile_position):

        tile_fgnd_frac = _compute_tile_foreground_fraction_single(
            slide_path, im_fgnd_mask_lres, fgnd_seg_scale,
            it_kwargs, tile_position, style=style,
        )

    else:

        msg = 'Invalid value for tile_position. Must be None or int'
        raise ValueError(
            msg)

    return tile_fgnd_frac


def _compute_tile_foreground_fraction_single(slide_path, im_fgnd_mask_lres,
                                             fgnd_seg_scale, it_kwargs,
                                             tile_position, style=None):

    # get slide tile source
    ts = large_image.getTileSource(slide_path, style=style)

    # get requested tile
    tile = ts.getSingleTile(tile_position=tile_position,
                            format=large_image.tilesource.TILE_FORMAT_NUMPY,
                            **it_kwargs)

    # get current region in base_pixels
    rgn_hres = {'left': tile['gx'], 'top': tile['gy'],
                'right': tile['gx'] + tile['gwidth'],
                'bottom': tile['gy'] + tile['gheight'],
                'units': 'base_pixels'}

    # get foreground mask for current tile at low resolution
    rgn_lres = ts.convertRegionScale(rgn_hres,
                                     targetScale=fgnd_seg_scale,
                                     targetUnits='mag_pixels')

    top = int(rgn_lres['top'])
    bottom = int(rgn_lres['bottom'])
    left = int(rgn_lres['left'])
    right = int(rgn_lres['right'])

    im_tile_fgnd_mask_lres = im_fgnd_mask_lres[top:bottom, left:right]

    # compute foreground fraction
    tile_fgnd_frac = im_tile_fgnd_mask_lres.mean()

    if np.isnan(tile_fgnd_frac):
        tile_fgnd_frac = 0

    return tile_fgnd_frac



import numpy as np


def eigen(im_hess):
    """
    Calculates the eigenvectors of the hessian volumes 'H'
    generated by Hessian.py

    Parameters
    ----------
    im_hess : array_like
        M x N x 4 hessian matrix - H[:,:,0] = dxx,
        H[:,:,1] = H[:,:,2] = dxy, H[:,:,3] = dyy.

    Returns
    -------
    lambda_ : array_like
        M x N x 2 image of eigenvalues.
    v1 : array_like
        M x N x 2 eigenvector for lambda_(:,:,0)
    v2 :
        M x N x 2 eigenvector for lambda_(:,:,1)

    """
    # get size of H
    sizeX = im_hess.shape[0]
    sizeY = im_hess.shape[1]

    # initialize lambda_, v1 and v2
    lambda_ = np.zeros((sizeX, sizeY, 2))
    v1 = np.zeros((sizeX, sizeY, 2))
    v2 = np.zeros((sizeX, sizeY, 2))

    # compute eigenvalues of H
    radical = np.sqrt((im_hess[:, :, 0] - im_hess[:, :, 3]) ** 2 +
                      4 * im_hess[:, :, 1] ** 2)
    lambda_[:, :, 0] = (im_hess[:, :, 0] + im_hess[:, :, 3] + radical) / 2
    lambda_[:, :, 1] = (im_hess[:, :, 0] + im_hess[:, :, 3] - radical) / 2

    # compute eigenvectors of H
    v1[:, :, 0] = 2 * im_hess[:, :, 1]
    v1[:, :, 1] = im_hess[:, :, 3] - im_hess[:, :, 0] + radical
    norms = np.sqrt(v1[:, :, 0]**2 + v1[:, :, 1]**2)

    # normalize eigenvectors of H
    with np.errstate(divide='ignore', invalid='ignore'):
        v1[:, :, 0] = np.true_divide(v1[:, :, 0], norms)
        v1[:, :, 1] = np.true_divide(v1[:, :, 1], norms)
        # check -inf inf NaN
        v1[:, :, 0][~np.isfinite(v1[:, :, 0])] = 0
        v1[:, :, 1][~np.isfinite(v1[:, :, 1])] = 0
    v2[:, :, 0] = -v1[:, :, 1]
    v2[:, :, 1] = v1[:, :, 0]

    # order by magnitude
    swap = np.where(abs(lambda_[:, :, 0]) > abs(lambda_[:, :, 1]))

    # swap between lambda_[:, :, 0] and lambda_[:, :, 1]
    lambda_[:, :, 0][swap], lambda_[:, :, 1][swap] = \
        lambda_[:, :, 1][swap], lambda_[:, :, 0][swap]

    # swap between v1 and v2
    v1[:, :, 0][swap], v2[:, :, 0][swap] = \
        v2[:, :, 0][swap], v1[:, :, 0][swap]

    v1[:, :, 1][swap], v2[:, :, 1][swap] = \
        v2[:, :, 1][swap], v1[:, :, 1][swap]

    return lambda_, v1, v2



import numpy as np

from .condense import condense


def area_open(im_label, min_area):
    """Removes small objects from label image.

    Parameters
    ----------
    im_label : array_like
        A uint32 type label image generated by segmentation methods.
    min_area : int
        minimum area threshold for objects. Objects with fewer than 'min_area'
        pixels will be zeroed to merge with background.

    Returns
    -------
    im_open : array_like
        A uint32 label where objects with pixels < min_area are removed.

    Notes
    -----
    Objects are assumed to have positive nonzero values. im_label image will be
    condensed during processing.

    See Also
    --------
    histomicstk.segmentation.label.condense,
    histomicstk.segmentation.label.shuffle,
    histomicstk.segmentation.label.split,
    histomicstk.segmentation.label.width_open

    """
    import scipy.ndimage as ndi

    # copy input image
    im_open = im_label.copy()

    # condense label image
    if np.unique(im_open).size - 1 != im_open.max():
        im_open = condense(im_open)

    # count pixels in each object
    Counts, Edges = np.histogram(im_open, bins=im_open.max() + 1)

    # get locations of objects in initial image
    Locations = ndi.find_objects(im_open)

    # iterate through objects, zeroing where needed
    for i in np.arange(1, Counts.size):
        if Counts[i] < min_area:

            # extract object from label image
            Template = im_open[Locations[i - 1]]

            # label mask of object 'i'
            Template[Template == i] = 0

    # condense to fill gaps
    im_open = condense(im_open)

    return im_open



from collections import namedtuple

import numpy as np
from numpy import linalg

PopStats = namedtuple('PopStats', ['mean', 'stddev', 'min_max_ratio', 'disorder'])
PolyProps = namedtuple('PolyProps', ['area', 'peri', 'max_dist'])
TriProps = namedtuple('TriProps', ['sides', 'area'])
DensityProps = namedtuple('DensityProps', ['neighbors_in_distance',
                                           'distance_for_neighbors'])
Props = namedtuple('Props', ['voronoi', 'delaunay', 'mst_branches', 'density'])


def compute_global_cell_graph_features(
        centroids,
        neighbor_distances=None,
        neighbor_counts=(3, 5, 7),
):
    r"""Compute global (i.e., not per-nucleus) features of the nuclei with
    the given centroids based on the partitioning of the space into
    Voronoi cells and on the induced graph structure.

    Parameters
    ----------
    centroids : array_like
        Nx2 numpy array of nuclear centroids
    neighbor_distances : array_like
        Radii to count neighbors in
    neighbor_counts : sequence
        Sequence of numbers of neighbors, each of which is used to
        compute statistics relating to the distance required to reach
        that many neighbors.

    Returns
    -------
    props : pandas.DataFrame
        A single-row DataFrame with the following columns:

        - voronoi\_...: Voronoi diagram features

          - area\_...: Polygon area features
          - peri\_...: Polygon perimeter features
          - max_dist\_...: Maximum distance in polygon features

        - delaunay\_...: Delaunay triangulation features

          - sides\_...: Triangle side length features
          - area\_...: Triangle area features

        - mst_branches\_...: Minimum spanning tree branch features
        - density\_...: Density features

          - neighbors_in_distance\_...

            - 0, 1, ..., len(neighbor_distances) - 1: Neighbor count
              within given radius features.

          - distance_for_neighbors\_...

            - 0, 1, ..., len(neighbor_counts) - 1: Minimum distance to
              enclose count neighbors features

        The "..."s are meant to signify that what precedes is the
        start of a column name.  At the end of each column name is one
        of 'mean', 'stddev', 'min_max_ratio', and 'disorder'.
        'min_max_ratio' is the minimum-to-maximum ratio, and disorder
        is stddev / (mean + stddev).

    Note
    ----
    The indices for the density features are with respect to the
    *sorted* values of the corresponding argument sequence.

    References
    ----------
    .. [#] Doyle, S., Agner, S., Madabhushi, A., Feldman, M., & Tomaszewski, J.
       (2008, May).  Automated grading of breast cancer histopathology using
       spectral clustering with textural and architectural image features.
       In Biomedical Imaging: From Nano to Macro, 2008.  ISBI 2008.
       5th IEEE International Symposium on (pp. 496-499).  IEEE.

    """
    if neighbor_distances is None:
        neighbor_distances = 10. * np.arange(1, 6)
    return _flatten_to_dataframe(_compute_global_cell_graph_features(
        centroids,
        neighbor_distances,
        neighbor_counts,
    ))


def _compute_global_cell_graph_features(
        centroids,
        neighbor_distances,
        neighbor_counts,
):
    """Internal support for compute_global_cell_graph_features that
    returns its result in a nested nametuple structure instead of a
    pandas DataFrame.

    """
    from scipy import sparse
    from scipy.sparse.csgraph import minimum_spanning_tree
    from scipy.spatial import Voronoi
    from scipy.spatial import cKDTree as KDTree
    from scipy.spatial.distance import pdist

    vor = Voronoi(centroids)
    centroids = vor.points
    vertices = vor.vertices

    regions = [r for r in vor.regions if r and -1 not in r]
    areas = np.array([_poly_area(vertices[r]) for r in regions])
    peris = np.array([_poly_peri(vertices[r]) for r in regions])
    max_dists = np.array([pdist(vertices[r]).max() for r in regions])
    poly_props = PolyProps._make(map(_pop_stats, (areas, peris, max_dists)))

    # Assume that each Voronoi vertex is on exactly three ridges.
    ridge_points = vor.ridge_points
    # This isn't exactly the collection of sides, since if they should
    # be counted per-triangle then we weight border ridges wrong
    # relative to ridges that are part of two triangles.
    ridge_lengths = _dist(*np.swapaxes(centroids[ridge_points], 0, 1))
    sides = ridge_lengths
    # Point indices of each triangle
    tris = [[] for _ in vertices]
    for p, ri in enumerate(vor.point_region):
        for vi in vor.regions[ri]:
            if vi != -1:
                tris[vi].append(p)
    # This will only fail in particular symmetrical cases where a
    # Voronoi vertex is associated with more than three centroids.
    # Since this should be unlikely in practice, we don't handle it
    # beyond throwing an AssertionError
    assert all(len(t) == 3 for t in tris)
    tris = np.asarray(tris)
    areas = np.array([_poly_area(centroids[t]) for t in tris])
    tri_props = TriProps._make(map(_pop_stats, (sides, areas)))

    graph = sparse.coo_matrix((ridge_lengths, np.sort(ridge_points).T),
                              (len(centroids), len(centroids)))
    mst = minimum_spanning_tree(graph)
    # Without looking into exactly how minimum_spanning_tree
    # constructs its output, elimate any explicit zeros to be on the
    # safe side.
    mst_branches = _pop_stats(mst.data[mst.data != 0])

    tree = KDTree(centroids)
    neighbors_in_distance = {
        # Yes, we just throw away the actual points
        r: _pop_stats(np.stack(np.array(list(map(len, tree.query_ball_tree(tree, r))))) - 1)
        for r in neighbor_distances
    }

    distance_for_neighbors = dict(zip(
        neighbor_counts,
        map(_pop_stats, tree.query(centroids, [c + 1 for c in neighbor_counts])[0].T),
    ))
    density_props = DensityProps(neighbors_in_distance, distance_for_neighbors)

    return Props(poly_props, tri_props, mst_branches, density_props)


def _poly_area(vertices):
    return abs(_poly_signed_area(vertices))


def _poly_signed_area(vertices):
    return .5 * linalg.det(
        np.stack((vertices, np.roll(vertices, -1, axis=-2)), -1),
    ).sum(-1)


def _poly_peri(vertices):
    return _dist(vertices, np.roll(vertices, -1, axis=-2)).sum(-1)


def _dist(x, y):
    """Compute the distance between two sets of points.  Has signature
    (i),(i)->().

    """
    return (np.subtract(x, y) ** 2).sum(-1) ** .5


def _pop_stats(pop):
    # Filter out outliers (here defined as points more than three
    # standard deviations away from the mean)
    while True:
        mean = pop.mean()
        stddev = pop.std()
        mask = abs(pop - mean) <= 3 * stddev
        if mask.all():
            break
        pop = pop[mask]
    if pop.max():
        minmaxr = pop.min() / pop.max()
    else:
        minmaxr = 0.0
    if mean + stddev:
        disorder = stddev / (mean + stddev)
    else:
        disorder = 0.0
    return PopStats(mean, stddev, minmaxr, disorder)


def _flatten_to_dataframe(nt):
    """Flatten the result of _compute_global_cell_graph_features to the
    DataFrame returned by compute_global_cell_graph_features.

    """
    from pandas import DataFrame

    return DataFrame(_flatten_to_dict(nt), index=[0])


def _flatten_to_dict(nt, prefix=''):
    result = {}
    assert isinstance(nt, (tuple, dict))
    if isinstance(nt, tuple):
        d = nt._asdict()
    else:  # nt is a dict
        # We only have numeric keys, and they may be non-nice floats,
        # so just number them by sort order instead of doing something
        # else for names
        d = {str(i): kv[1] for i, kv in enumerate(sorted(nt.items()))}
    for k, v in d.items():
        if not isinstance(v, (tuple, dict)):
            # Terminate
            result[prefix + k] = v
        else:
            result.update(_flatten_to_dict(v, prefix + k + '_'))
    return result



"""Placeholder."""
import numpy as np

import histomicstk.utils as utils
from histomicstk.preprocessing import color_conversion


def color_convolution(im_stains, w, I_0=None):
    """Perform Color Convolution.

    Reconstructs a color image from the stain matrix `w` and
    the individual images stored as channels in `im_stains` and generated
    by ColorDeconvolution.

    Parameters
    ----------
    im_stains : array_like
        An RGB image where in each channel contains image of one stain
    w : array_like
        A 3x3 matrix containing the stain colors in its columns.
        In the case of two stains, the third column is zero and will be
        complemented using cross-product. The matrix should contain a
        minimum two nonzero columns.
    I_0 : float or array_like, optional
        A float a 3-vector containing background RGB intensities.
        If unspecified, use the old OD conversion.

    Returns
    -------
    im_rgb : array_like
        Reconstructed RGB image with intensity values ranging from [0, 255],
        suitable for display.

    See Also
    --------
    histomicstk.preprocessing.color_deconvolution.complement_stain_matrix,
    histomicstk.preprocessing.color_deconvolution.color_deconvolution
    histomicstk.preprocessing.color_conversion.rgb_to_od
    histomicstk.preprocessing.color_conversion.od_to_rgb
    histomicstk.preprocessing.color_conversion.rgb_to_sda
    histomicstk.preprocessing.color_conversion.sda_to_rgb

    """
    # transform 3D input stain image to 2D stain matrix format
    m = utils.convert_image_to_matrix(im_stains)

    # transform input stains to optical density values, convolve and
    # tfm back to stain
    sda_fwd = color_conversion.rgb_to_sda(m, 255 if I_0 is not None else None,
                                          allow_negatives=True)
    sda_conv = np.dot(w, sda_fwd)
    sda_inv = color_conversion.sda_to_rgb(sda_conv, I_0)

    # reshape output, transform type
    im_rgb = (utils.convert_matrix_to_image(sda_inv, im_stains.shape)
              .clip(0, 255).astype(np.uint8))

    return im_rgb



import numpy as np


def rag(im_label, neigh_conn=4):
    """Constructs a region adjacency graph for a label image using either
    4-neighbor or 8-neighbor connectivity. Background pixels are not included
    (im_label == 0). Not intended to build large graphs from individual pixels.

    Parameters
    ----------
    im_label : array_like
        im_label image where positive values (im_label > 0) correspond to foreground
        objects of interest.

    neigh_conn : float
        The neighbor connectivity to use, either '4' or '8'. Default value = 4.

    Returns
    -------
    adj_mat : array_like
        A binary matrix of size N x N, where N is the number of objects in
        im_label. A value of 'True' at adj_mat(i,j) indicates that objects 'i'
        and 'j' are neigh_conn.

    """
    # initialize adjacency matrix for im_label.max() total regions
    adj_mat = np.zeros((im_label.max(), im_label.max()), dtype=bool)

    # process 4-neigh_conn horizontal connections
    Mask = (im_label[:, 1:-1] != im_label[:, 0:-2]) & \
           (im_label[:, 1:-1] != 0) & (im_label[:, 0:-2] != 0)
    Indices = np.nonzero(Mask)
    Xc = np.concatenate((im_label[Indices[0], Indices[1]][:, np.newaxis],
                         im_label[Indices[0], Indices[1] + 1][:, np.newaxis]),
                        axis=1)

    # process 4-neigh_conn vertical connections
    Mask = (im_label[1:-1, :] != im_label[0:-2, :]) & \
           (im_label[1:-1, :] != 0) & (im_label[0:-2, :] != 0)
    Indices = np.nonzero(Mask)
    Xc = np.concatenate((Xc,
                         np.concatenate((im_label[Indices[0], Indices[1]]
                                         [:, np.newaxis],
                                         im_label[Indices[0] + 1, Indices[1]]
                                         [:, np.newaxis]), axis=1)),
                        axis=0)

    # process additional 8-neighbor relationships
    if neigh_conn == 8:

        # shift upper-right ([1, 1])
        Mask = (im_label[1:-1, 0:-2] != im_label[0:-2, 1:-1]) & \
               (im_label[1:-1, 0:-2] != 0) & (im_label[0:-2, 1:-1] != 0)
        Indices = np.nonzero(Mask)
        Xc = np.concatenate((
            Xc,
            np.concatenate((
                im_label[Indices[0] + 1, Indices[1]][:, np.newaxis],
                im_label[Indices[0], Indices[1] + 1][:, np.newaxis]),
                axis=1)),
            axis=0,
        )

        # shift upper-left ([-1, 1])
        Mask = (im_label[0:-2, 0:-2] != im_label[1:-1, 1:-1]) & \
               (im_label[0:-2, 0:-2] != 0) & (im_label[1:-1, 1:-1] != 0)
        Indices = np.nonzero(Mask)
        Xc = np.concatenate((
            Xc,
            np.concatenate((
                im_label[Indices[0], Indices[1]][:, np.newaxis],
                im_label[Indices[0] + 1, Indices[1] + 1][:, np.newaxis]),
                axis=1)),
            axis=0,
        )

    # add entries to adjacency matrix
    for i in range(Xc.shape[0]):
        adj_mat[Xc[i, 0] - 1, Xc[i, 1] - 1] = True
        adj_mat[Xc[i, 1] - 1, Xc[i, 0] - 1] = True

    # return result
    return adj_mat



import numpy as np

import histomicstk as htk
import histomicstk.filters.shape as htk_shape_filters


def detect_nuclei_kofahi(im_nuclei_stain, im_nuclei_fgnd_mask, min_radius,
                         max_radius, min_nucleus_area, local_max_search_radius):
    """Performs a nuclear segmentation using kofahi's method.

    This method uses scale-adaptive multi-scale Laplacian-of-Gaussian filtering
    for blob enhancement and a local maximum clustering for segmentation. The
    original implementation described by Al-Kofahi et al. uses Laplacian of Gaussian
    but this function replaces it with Difference of Gaussian to improve speed.

    Parameters
    ----------
    im_nuclei_stain : array_like
        A hematoxylin intensity image obtained from ColorDeconvolution.
    im_nuclei_fgnd_mask: array_like
        A binary mask of the nuclear foreground typically obtained by applying
        a threshold on the hematoxylin/nuclei stain image
    min_radius : float
        Minimum nuclear radius (used to set min sigma of the multiscale LoG filter)
    max_radius : float
        Maximum nuclear radius (used to set max sigma of the multiscale LoG filter)
    min_nucleus_area : int
        Minimum area that each nucleus should have
    local_max_search_radius : float
        Local max search radius used for detection seed points in nuclei

    Returns
    -------
    im_nuclei_seg_mask : array_like
        A 2D array mask of the nuclei segmentation.

    References
    ----------
    .. [#] Y. Al-Kofahi et al "Improved Automatic Detection and Segmentation of
       Cell Nuclei in Histopathology Images" in IEEE Transactions on Biomedical
       Engineering, Volume: 57, Issue: 4, doi: 10.1109/TBME.2009.2035102,
       April 2010.

    """
    import scipy as sp
    import skimage.morphology

    # smooth foreground mask with closing and opening
    im_nuclei_fgnd_mask = skimage.morphology.closing(
        im_nuclei_fgnd_mask, skimage.morphology.disk(3))

    im_nuclei_fgnd_mask = skimage.morphology.opening(
        im_nuclei_fgnd_mask, skimage.morphology.disk(3))

    im_nuclei_fgnd_mask = sp.ndimage.binary_fill_holes(
        im_nuclei_fgnd_mask)

    if not np.any(im_nuclei_fgnd_mask):
        return im_nuclei_fgnd_mask

    # run adaptive multi-scale LoG filter
    im_log_max, im_sigma_max = htk_shape_filters.cdog(
        im_nuclei_stain, im_nuclei_fgnd_mask,
        sigma_min=min_radius / np.sqrt(2),
        sigma_max=max_radius / np.sqrt(2),
    )

    # apply local maximum clustering
    im_nuclei_seg_mask, seeds, maxima = htk.segmentation.nuclear.max_clustering(
        im_log_max, im_nuclei_fgnd_mask, local_max_search_radius)

    if seeds is None:
        return im_nuclei_seg_mask

    # split any objects with disconnected fragments
    im_nuclei_seg_mask = htk.segmentation.label.split(im_nuclei_seg_mask,
                                                      conn=8)

    # filter out small objects
    im_nuclei_seg_mask = htk.segmentation.label.area_open(
        im_nuclei_seg_mask, min_nucleus_area).astype(int)

    return im_nuclei_seg_mask



import json
import logging
import os
import time
from pathlib import Path

import large_image
import numpy as np

import histomicstk
import histomicstk.preprocessing.color_normalization as htk_cnorm
import histomicstk.segmentation.label as htk_seg_label
import histomicstk.segmentation.nuclear as htk_nuclear
import histomicstk.utils as htk_utils
from histomicstk.cli import utils as cli_utils
from histomicstk.cli.utils import CLIArgumentParser

logging.basicConfig(level=logging.CRITICAL)


def check_args(args):

    if not os.path.isfile(args.inputImageFile):
        msg = 'Input image file does not exist.'
        raise OSError(msg)

    if len(args.reference_mu_lab) != 3:
        msg = 'Reference Mean LAB should be a 3 element vector.'
        raise ValueError(msg)

    if len(args.reference_std_lab) != 3:
        msg = 'Reference Stddev LAB should be a 3 element vector.'
        raise ValueError(msg)

    if len(args.analysis_roi) != 4:
        msg = 'Analysis ROI must be a vector of 4 elements.'
        raise ValueError(msg)

    if os.path.splitext(args.outputNucleiFeatureFile)[1] not in ['.csv', '.h5']:
        msg = 'Extension of output feature file must be .csv or .h5'
        raise ValueError(msg)


def main(args):  # noqa
    import dask
    import pandas as pd

    total_start_time = time.time()

    print('\n>> CLI Parameters ...\n')

    print(args)

    check_args(args)

    feature_file_format = os.path.splitext(args.outputNucleiFeatureFile)[1]

    if np.all(np.array(args.analysis_roi) == -1):
        process_whole_image = True
    else:
        process_whole_image = False

    # Provide default value for tile_overlap
    tile_overlap = args.tile_overlap_value
    if tile_overlap == -1:
        tile_overlap = (args.max_radius + 1) * 4

    #
    # Initiate Dask client
    #
    print('\n>> Creating Dask client ...\n')

    start_time = time.time()

    c = cli_utils.create_dask_client(args)

    print(c)

    dask_setup_time = time.time() - start_time
    print(f'Dask setup time = {dask_setup_time} seconds')

    #
    # Read Input Image
    #
    print('\n>> Reading input image ... \n')

    ts = large_image.getTileSource(args.inputImageFile)

    ts_metadata = ts.getMetadata()

    print(json.dumps(ts_metadata, indent=2))

    is_wsi = ts_metadata['magnification'] is not None

    #
    # Compute tissue/foreground mask at low-res for whole slide images
    #
    if is_wsi and process_whole_image:

        print('\n>> Computing tissue/foreground mask at low-res ...\n')

        start_time = time.time()

        im_fgnd_mask_lres, fgnd_seg_scale = \
            cli_utils.segment_wsi_foreground_at_low_res(ts)

        fgnd_time = time.time() - start_time

        print('low-res foreground mask computation time = {}'.format(
            cli_utils.disp_time_hms(fgnd_time)))

    #
    # Compute foreground fraction of tiles in parallel using Dask
    #
    tile_fgnd_frac_list = [1.0]

    it_kwargs = {
        'tile_size': {'width': args.analysis_tile_size},
        'scale': {'magnification': args.analysis_mag},
        'tile_overlap': {'x': tile_overlap, 'y': tile_overlap},
    }

    if not process_whole_image:

        it_kwargs['region'] = {
            'left': args.analysis_roi[0],
            'top': args.analysis_roi[1],
            'width': args.analysis_roi[2],
            'height': args.analysis_roi[3],
            'units': 'base_pixels',
        }

    if is_wsi:

        print('\n>> Computing foreground fraction of all tiles ...\n')

        start_time = time.time()

        num_tiles = ts.getSingleTile(**it_kwargs)['iterator_range']['position']

        print(f'Number of tiles = {num_tiles}')

        if process_whole_image:

            tile_fgnd_frac_list = htk_utils.compute_tile_foreground_fraction(
                args.inputImageFile, im_fgnd_mask_lres, fgnd_seg_scale,
                it_kwargs,
            )

        else:

            tile_fgnd_frac_list = np.full(num_tiles, 1.0)

        num_fgnd_tiles = np.count_nonzero(
            tile_fgnd_frac_list >= args.min_fgnd_frac)

        percent_fgnd_tiles = 100.0 * num_fgnd_tiles / num_tiles

        fgnd_frac_comp_time = time.time() - start_time

        print('Number of foreground tiles = {:d} ({:.2f}%)'.format(
            num_fgnd_tiles, percent_fgnd_tiles))

        print('Tile foreground fraction computation time = {}'.format(
            cli_utils.disp_time_hms(fgnd_frac_comp_time)))

    #
    # Compute reinhard stats for color normalization
    #
    src_mu_lab = None
    src_sigma_lab = None

    if is_wsi and process_whole_image:

        print('\n>> Computing reinhard color normalization stats ...\n')

        start_time = time.time()

        src_mu_lab, src_sigma_lab = htk_cnorm.reinhard_stats(
            args.inputImageFile, 0.01, magnification=args.analysis_mag)

        rstats_time = time.time() - start_time

        print('Reinhard stats computation time = {}'.format(
            cli_utils.disp_time_hms(rstats_time)))

    #
    # Detect and compute nuclei features in parallel using Dask
    #
    print('\n>> Detecting nuclei and computing features ...\n')

    start_time = time.time()

    tile_result_list = []

    for tile in ts.tileIterator(**it_kwargs):

        tile_position = tile['tile_position']['position']

        if is_wsi and tile_fgnd_frac_list[tile_position] <= args.min_fgnd_frac:
            continue

        # detect nuclei
        cur_result = dask.delayed(htk_nuclear.detect_tile_nuclei)(
            tile,
            args,
            src_mu_lab, src_sigma_lab,
            return_fdata=True,
        )

        # append result to list
        tile_result_list.append(cur_result)

    tile_result_list = dask.delayed(tile_result_list).compute()

    nuclei_annot_list = [annot
                         for annot_list, fdata in tile_result_list
                         for annot in annot_list]

    nuclei_fdata = pd.DataFrame()

    if len(nuclei_annot_list) > 0:

        nuclei_fdata = pd.concat([
            fdata
            for annot_list, fdata in tile_result_list if fdata is not None],
            ignore_index=True,
        )

    nuclei_detection_time = time.time() - start_time

    print(f'Number of nuclei = {len(nuclei_annot_list)}')
    print('Nuclei detection time = {}'.format(
        cli_utils.disp_time_hms(nuclei_detection_time)))

    #
    # Remove overlapping nuclei
    #
    if args.remove_overlapping_nuclei_segmentation:
        print('\n>> Removing overlapping nuclei segmentations ...\n')
        nuclei_removal_start_time = time.time()

        nuclei_annot_list, selected_nuclei = htk_seg_label.remove_overlap_nuclei(
            nuclei_annot_list, args.nuclei_annotation_format, return_selected_nuclei=True)
        nuclei_removal_setup_time = time.time() - nuclei_removal_start_time

        # Remove features for overlapping nuclei
        nuclei_fdata = nuclei_fdata[nuclei_fdata.index.isin(selected_nuclei)]

        print(f'Number of nuclei after overlap removal {len(nuclei_annot_list)}')
        print('Nuclei removal processing time = {}'.format(
            cli_utils.disp_time_hms(nuclei_removal_setup_time)))

    #
    # Write annotation file
    #
    print('\n>> Writing annotation file ...\n')

    annot_fname = os.path.splitext(
        os.path.basename(args.outputNucleiAnnotationFile))[0]

    if args.in_annotations:
        for key in args.in_annotations.split(','):
            if key in nuclei_fdata:
                for row, value in zip(nuclei_annot_list, nuclei_fdata[key]):
                    if value is not None:
                        row.setdefault('user', {})[key] = value
    annotation = {
        'name': annot_fname + '-nuclei-' + args.nuclei_annotation_format,
        'elements': nuclei_annot_list,
        'attributes': {
            'src_mu_lab': None if src_mu_lab is None else src_mu_lab.tolist(),
            'src_sigma_lab': None if src_sigma_lab is None else src_sigma_lab.tolist(),
            'params': vars(args),
            'cli': Path(__file__).stem,
            'version': histomicstk.__version__},
    }

    with open(args.outputNucleiAnnotationFile, 'w') as annotation_file:
        json.dump(annotation, annotation_file, separators=(',', ':'), sort_keys=False)

    #
    # Create CSV Feature file
    #
    print('>> Writing CSV feature file')

    if feature_file_format == '.csv':

        nuclei_fdata.to_csv(args.outputNucleiFeatureFile, index=False)

    elif feature_file_format == '.h5':

        nuclei_fdata.to_hdf(args.outputNucleiFeatureFile, 'Features',
                            format='table', mode='w')

    else:

        msg = 'Extension of output feature file must be .csv or .h5'
        raise ValueError(msg)

    total_time_taken = time.time() - total_start_time

    print('Total analysis time = {}'.format(
        cli_utils.disp_time_hms(total_time_taken)))


if __name__ == '__main__':
    main(CLIArgumentParser().parse_args())



"""
This package contains functions for segmenting a variety of objects/structures
(e.g. nuclei, tissue, cytoplasm) found in histopathology images.
"""

# make functions available at the package level using shadow imports
# since we mostly have one function per file
from histomicstk.utils.simple_mask import simple_mask

# import sub-packages to support nested calls
from . import label, level_set, nuclear, positive_pixel_count
from .embed_boundaries import embed_boundaries
from .rag import rag
from .rag_add_layer import rag_add_layer
from .rag_color import rag_color

# list out things that are available for public use
__all__ = (

    # functions and classes of this package
    'embed_boundaries',
    'rag_color',
    'rag',
    'rag_add_layer',
    'simple_mask',

    # sub-packages
    'label',
    'level_set',
    'nuclear',
    'positive_pixel_count',
)



import numpy as np

import histomicstk.utils as htk_utils


def vesselness(im_input, sigma):
    """
    Calculates vesselness measure for grayscale image `im_input` at scale `sigma`.
    Also returns eigenvalues and vectors used for vessel salience filters.

    Parameters
    ----------
    im_input : array_like
        M x N grayscale image.
    sigma : double
        standard deviation of gaussian kernel.

    Returns
    -------
    Deviation : array_like
        M x N image of deviation from blob
    Frobenius : array_like
        M x N image of frobenius norm of Hessian - measures presence of
        structure.
    E : array_like
        M x N x 2 eigenvalue image - see eigen.py.
    Theta : array_like
        M x N eigenvector angle image for E(:,:,0) in radians
        see eigen.py. Oriented parallel to vessel structures.

    References
    ----------
    .. [#] Frangi, Alejandro F., et al. "Multiscale vessel enhancement
       filtering." Medical Image Computing and Computer-Assisted
       Interventation. MICCAI98. Springer Berlin Heidelberg,1998. 130-137.

    """
    # calculate hessian matrix
    H = sigma ** 2 * htk_utils.hessian(im_input, sigma)

    # calculate eigenvalue image
    E, V1, V2 = htk_utils.eigen(H)

    # compute blobness measures
    Deviation = E[:, :, 0] / (E[:, :, 1] + np.spacing(1))
    Frobenius = np.sqrt(E[:, :, 0]**2 + E[:, :, 1]**2)

    # calculate angles for 'Theta'
    Theta = np.arctan2(V1[:, :, 1], V1[:, :, 0])

    return Deviation, Frobenius, E, Theta



import numpy as np

from . import _linalg as linalg


def find_stain_index(reference, w):
    """Identify the stain vector in w that best aligns with the reference vector.

    This is used with adaptive deconvolution routines where the order of returned stain
    vectors is not guaranteed. This function identifies the stain vector of w that most
    closely aligns with the provided reference.

    Parameters
    ----------
    reference : array_like
        1D array representing the stain vector query.
    w : array_like
        3xN array of where columns represent stain vectors to search.

    Returns
    -------
    i : int
        Column index of stain vector with best alignment to reference.

    Notes
    -----
    Vectors are normalized to unit-norm prior to comparison using dot product. Alignment
    is determined by vector angles and not distances.

    See Also
    --------
    histomicstk.preprocessing.color_deconvolution.separate_stains_macenko_pca
    histomicstk.preprocessing.color_deconvolution.color_deconvolution

    """
    dot_products = np.dot(
        linalg.normalize(np.array(reference)), linalg.normalize(np.array(w)),
    )
    return np.argmax(np.abs(dot_products))



import numpy as np


def compute_fsd_features(im_label, K=128, Fs=6, Delta=8, rprops=None):
    """
    Calculates `Fourier shape descriptors` for each objects.

    Parameters
    ----------
    im_label : array_like
        A labeled mask image wherein intensity of a pixel is the ID of the
        object it belongs to. Non-zero values are considered to be foreground
        objects.
    K : int, optional
        Number of points for boundary resampling to calculate fourier
        descriptors. Default value = 128.
    Fs : int, optional
        Number of frequency bins for calculating FSDs. Default value = 6.
    Delta : int, optional
        Used to dilate nuclei and define cytoplasm region. Default value = 8.

    rprops : output of skimage.measure.regionprops, optional
        rprops = skimage.measure.regionprops( im_label ). If rprops is not
        passed then it will be computed inside which will increase the
        computation time.

    Returns
    -------
    fdata: Pandas data frame containing the FSD features for each
           object/label.

    References
    ----------
    .. [#] D. Zhang et al. "A comparative study on shape retrieval using
       Fourier descriptors with different shape signatures," In Proc.
       ICIMADE01, 2001.

    """
    import pandas as pd
    from skimage.measure import regionprops

    from histomicstk.segmentation.label import trace_object_boundaries

    # List of feature names
    feature_list = []
    for i in range(Fs):
        feature_list = np.append(feature_list, 'Shape.FSD' + str(i + 1))

    # get Label size x
    sizex = im_label.shape[0]
    sizey = im_label.shape[1]

    # get the number of objects in Label
    if rprops is None:
        rprops = regionprops(im_label)

    # create pandas data frame containing the features for each object
    numFeatures = len(feature_list)
    numLabels = len(rprops)

    # pre-compute Interval outside the loop
    Interval = np.round(
        np.power(2, np.linspace(0, np.log2(K) - 1, Fs + 1, endpoint=True)),
    ).astype(np.uint8)

    # initialize an empty list to collect data
    data_list = []

    for i in range(numLabels):
        # get bounds of dilated nucleus
        min_row, max_row, min_col, max_col = _GetBounds(
            rprops[i].bbox,
            Delta,
            sizex,
            sizey,
        )

        # grab label mask
        lmask = (im_label[min_row:max_row, min_col:max_col] == rprops[i].label).astype(
            bool,
        )
        # trace_object_boundaries requires that that edge rows and columns be
        # false as it does no bounds checking
        if not min_row or not min_col or max_row + 1 == sizex or max_col + 1 == sizey:
            lmask = np.pad(lmask, (
                (1 if not min_row else 0, 1 if max_row + 1 == sizex else 0),
                (1 if not min_col else 0, 1 if max_col + 1 == sizey else 0)))
        # find boundaries
        Bounds = trace_object_boundaries(
            lmask,
            conn=8,
            trace_all=False,
            x_start=None,
            y_start=None,
            max_length=None,
            simplify_colinear_spurs=False,
            eps_colinear_area=0.01,
            region_props=None,
        )
        Bounds = np.stack([Bounds[0][0][:-1], Bounds[1][0][:-1]], axis=-1)
        # check length of boundaries
        if len(Bounds) < 2:
            data_list.append(np.zeros(numFeatures))
        else:
            # compute fourier descriptors and collect data
            data_list.append(_FSDs(Bounds[:, 0], Bounds[:, 1], K, Interval))

    # create DataFrame after the loop
    fdata = pd.DataFrame(data_list, columns=feature_list)

    return fdata


def _InterpolateArcLength(X, Y, K):
    """
    Resamples boundary points [X, Y] at L total equal arc-length locations.

    Parameters
    ----------
    X : array_like
        x points of boundaries
    Y : array_like
        y points of boundaries
    K : int
        Number of points for boundary resampling to calculate fourier
        descriptors. Default value = 128.

    Returns
    -------
    iX : array_like
        L-length vector of horizontal interpolated coordinates with equal
        arc-length spacing.
    iY : array_like
        L-length vector of vertical interpolated coordinates with equal
        arc-length spacing.

    """
    # generate spaced points 0, 1/k, 1
    interval = np.linspace(0, 1, K + 1)
    # get segment lengths
    slens = np.sqrt(np.diff(X) ** 2 + np.diff(Y) ** 2)
    # normalize to unit length
    slens = np.true_divide(slens, slens.sum())
    # calculate cumulative length along boundary
    cumulative = np.zeros(len(slens) + 1)
    cumulative[1:] = np.cumsum(slens)
    # place points in 'Interval' along boundary
    locations = np.digitize(interval, cumulative)
    # clip to ends
    locations[locations > len(slens)] = len(slens)
    # linear interpolation
    Lie = (interval - cumulative[locations - 1]) / slens[locations - 1]
    iX = X[locations - 1] + (X[locations] - X[locations - 1]) * Lie
    iY = Y[locations - 1] + (Y[locations] - Y[locations - 1]) * Lie

    return iX, iY


def _FSDs(X, Y, K, Intervals):
    """
    Calculated FSDs from boundary points X,Y. Boundaries are resampled to have
    K equally spaced points (arclength) around the shape. The curvature is
    calculated using the cumulative angular function, measuring the
    displacement of the tangent angle from the starting point of the boundary.
    The K-length fft of the cumulative angular function is calculated, and
    then the elements of 'F' are summed as the spectral energy over
    'Intervals'.

    Parameters
    ----------
    X : array_like
        x points of boundaries
    Y : array_like
        y points of boundaries
    K : int
        Number of points for boundary resampling to calculate fourier
        descriptors. Default value = 128.
    Intervals : array_like
        Intervals spaced evenly over 1:K/2.

    Returns
    -------
    F : array_like
        length(Intervals) vector containing spectral energy of
        cumulative angular function, summed over defined 'Intervals'.

    """
    # check input 'Intervals'
    if Intervals[0] != 1.0:
        Intervals = np.hstack((1.0, Intervals))
    if Intervals[-1] != (K / 2):
        Intervals = np.hstack((Intervals, float(K)))
    # get length of intervals
    L = len(Intervals)
    # initialize F
    F = np.zeros((L - 1,)).astype(float)
    # interpolate boundaries
    iX, iY = _InterpolateArcLength(X, Y, K)
    # check if iXY.iX is not empty
    if iX.size:
        # calculate curvature
        Curvature = np.arctan2(
            (iY[1:] - iY[:-1]),
            (iX[1:] - iX[:-1]),
        )
        # make curvature cumulative
        Curvature = Curvature - Curvature[0]
        # calculate FFT
        z = 1 * np.cos(Curvature) + 1j * np.sin(Curvature)
        fX = np.fft.fft(z).T
        # spectral energy
        fX = fX * fX.conj()
        fX = (fX / fX.sum()) if fX.sum() else fX
        # calculate 'F' values
        for i in range(L - 1):
            F[i] = np.round(
                fX[Intervals[i] - 1: Intervals[i + 1]].sum(),
                L,
            ).real.astype(float)

    return F


def _GetBounds(bbox, delta, M, N):
    """
    Returns bounds of object in global label image.

    Parameters
    ----------
    bbox : tuple
        Bounding box (min_row, min_col, max_row, max_col).
    delta : int
        Used to dilate nuclei and define cytoplasm region.
        Default value = 8.
    M : int
        X size of label image.
    N : int
        Y size of label image.

    Returns
    -------
    min_row : int
        Minimum row of the region bounds.
    max_row : int
        Maximum row of the region bounds.
    min_col : int
        Minimum column of the region bounds.
    max_col : int
        Maximum column of the region bounds.

    """
    min_row, min_col, max_row, max_col = bbox

    min_row_out = max(0, (min_row - delta))
    max_row_out = min(M - 1, (max_row + delta))
    min_col_out = max(0, (min_col - delta))
    max_col_out = min(N - 1, (max_col + delta))

    return min_row_out, max_row_out, min_col_out, max_col_out



import numpy as np


def dilate_xor(im_label, neigh_width=8):
    """Computes a label mask highlighting a ring-like neighborhood of each
    object or region in a given label mask

    Parameters
    ----------
    im_label : array_like
        A labeled mask image wherein intensity of a pixel is the ID of the
        object it belongs to. Non-zero values are considered to be foreground
        objects.

    neigh_width : float, optional
        The width of the ring-like neighborhood around each object.

    Returns
    -------
    im_neigh_label : array_like
        A labeled mask image highlighting pixels in a ring-like neighborhood of
        width upto `neigh_width` around each object in the given label mask.
        The intensity of each pixel in the ring-like neighborhood is set
        equal to the label of the closest object in the given label mask.
        other pixels (including the ones inside objects) are set to zero.

    """
    from scipy.ndimage import distance_transform_edt

    # For each background pixel compute the distance to the nearest object and
    # the indices of the nearest object pixel
    im_dist, closest_obj_ind = distance_transform_edt(im_label == 0,
                                                      return_indices=True)
    closest_obj_rind, closest_obj_cind = closest_obj_ind

    # Get indices of background pixels within a given distance from an object
    neigh_rind, neigh_cind = np.where(
        np.logical_and(im_dist > 0, im_dist <= neigh_width),
    )

    # generate labeled neighborhood mask
    im_neigh_label = np.zeros_like(im_label)

    im_neigh_label[neigh_rind, neigh_cind] = im_label[
        closest_obj_rind[neigh_rind, neigh_cind],
        closest_obj_cind[neigh_rind, neigh_cind]]

    return im_neigh_label



"""
Created on Tue Aug 20 18:13:37 2019.

@author: tageldim
"""
import os
import sys

import pandas as pd
import pytest
from imageio.v2 import imread
from pandas import read_csv

from histomicstk.annotations_and_masks.annotation_and_mask_utils import \
    delete_annotations_in_slide
from histomicstk.annotations_and_masks.masks_to_annotations_handler import (
    get_annotation_documents_from_contours, get_contours_from_mask)

thisDir = os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0, os.path.join(thisDir, '../../../tests'))
from tests.htk_test_utilities import getTestFilePath, girderClient  # noqa


class TestMasksToAnnotations:
    """Test methods for getting ROI mask from annotations."""

    def _setup(self):
        # read GTCodes dataframe
        gtcodePath = getTestFilePath('sample_GTcodes.csv')
        self.GTCodes_df = read_csv(gtcodePath)
        self.GTCodes_df.index = self.GTCodes_df.loc[:, 'group']

        # read sample contours_df dataframe to test against
        contoursDfPath = getTestFilePath(os.path.join(
            'annotations_and_masks', 'sample_contours_df.tsv'))
        self.CONTOURS_DF = read_csv(contoursDfPath, sep='\t', index_col=0)

        # read mask
        self.X_OFFSET = 59206
        self.Y_OFFSET = 33505
        self.MASKNAME = (
            'TCGA-A2-A0YE-01Z-00-DX1.8A2E3094-5755-42BC-969D-7F0A2ECA0F39_'
            'left-%d_top-%d_mag-BASE.png' % (self.X_OFFSET, self.Y_OFFSET))
        MASKPATH = getTestFilePath(os.path.join(
            'annotations_and_masks', self.MASKNAME))
        self.MASK = imread(MASKPATH)

    def test_get_contours_from_mask(self):
        """Test get_contours_from_mask()."""
        self._setup()
        # get contours from mask
        # groups_to_get = [
        #     'mostly_tumor', 'mostly_stroma']
        groups_to_get = None
        contours_df = get_contours_from_mask(
            MASK=self.MASK, GTCodes_df=self.GTCodes_df,
            groups_to_get=groups_to_get,
            get_roi_contour=True, roi_group='roi',
            discard_nonenclosed_background=True,
            background_group='mostly_stroma',
            MIN_SIZE=30, MAX_SIZE=None, verbose=False,
            monitorPrefix=self.MASKNAME[:12] + ': getting contours')

        # make sure it is what we expect
        assert set(contours_df.columns) == set(self.CONTOURS_DF.columns)
        assert all(contours_df.iloc[:10, :] == self.CONTOURS_DF.iloc[:10, :])

    def test_get_contours_from_mask_with_groups(self):
        """Test get_contours_from_mask()."""
        self._setup()
        # get contours from mask
        groups_to_get = [
            'mostly_tumor', 'mostly_stroma']
        contours_df = get_contours_from_mask(
            MASK=self.MASK, GTCodes_df=self.GTCodes_df,
            groups_to_get=groups_to_get,
            get_roi_contour=True, roi_group='roi',
            discard_nonenclosed_background=True,
            background_group='mostly_stroma',
            MIN_SIZE=30, MAX_SIZE=None, verbose=False,
            monitorPrefix=self.MASKNAME[:12] + ': getting contours')

        # make sure it is what we expect
        assert set(contours_df.columns) == set(self.CONTOURS_DF.columns)
        assert all(contours_df.iloc[:10, :] == self.CONTOURS_DF.iloc[:10, :])
        assert len(contours_df) == 26

    def test_get_contours_from_mask_with_zeroes(self):
        """Test get_contours_from_mask()."""
        self._setup()
        groups_to_get = None
        gtcodes = pd.concat([self.GTCodes_df, pd.DataFrame([{
            'group': 'zeroes',
            'overlay_order': 4,
            'GT_code': 0,
            'is_roi': 0,
            'is_background_class': 0,
            'color': 'rgb(0,128,0)',
            'comments': 'zeroes'}])], ignore_index=True)
        gtcodes.index = gtcodes.loc[:, 'group']
        contours_df = get_contours_from_mask(
            MASK=self.MASK, GTCodes_df=gtcodes,
            groups_to_get=groups_to_get,
            get_roi_contour=True, roi_group='roi',
            discard_nonenclosed_background=True,
            background_group='mostly_stroma',
            MIN_SIZE=30, MAX_SIZE=None)

        # make sure it is what we expect
        assert set(contours_df.columns) == set(self.CONTOURS_DF.columns)
        assert all(contours_df.iloc[:10, :] == self.CONTOURS_DF.iloc[:10, :])
        assert len(contours_df) == 49

    @pytest.mark.usefixtures('girderClient')  # noqa
    def test_get_annotation_documents_from_contours(self, girderClient):  # noqa
        """Test get_contours_from_bin_mask()."""
        self._setup()
        original_iteminfo = girderClient.get(
            '/item', parameters={'text': 'TCGA-A2-A0YE-01Z-00-DX1'})[0]

        folder = girderClient.post(
            '/folder', data={
                'parentId': original_iteminfo['folderId'],
                'name': 'test-masks-annot-handler',
            })

        # copy the item
        sampleSlideItem = girderClient.post(
            '/item/%s/copy' % original_iteminfo['_id'], data={
                'name': 'TCGA-A2-A0YE-01Z.svs',
                'copyAnnotations': True,
                'folderId': folder['_id'],
            })
        sampleSlideId = str(sampleSlideItem['_id'])
        # get list of annotation documents
        annprops = {
            'X_OFFSET': self.X_OFFSET,
            'Y_OFFSET': self.Y_OFFSET,
            'opacity': 0.2,
            'lineWidth': 4.0,
        }
        annotation_docs = get_annotation_documents_from_contours(
            self.CONTOURS_DF.copy(), separate_docs_by_group=True,
            annots_per_doc=10, docnamePrefix='test', annprops=annprops,
            verbose=False,
            monitorPrefix=self.MASKNAME[:12] + ': annotation docs')

        # make sure its what we expect
        assert len(annotation_docs) == 8
        assert {j['name'] for j in annotation_docs} == {
            'test_blood_vessel-0',
            'test_exclude-0',
            'test_mostly_lymphocytic_infiltrate-0',
            'test_mostly_stroma-0',
            'test_mostly_tumor-0',
            'test_mostly_tumor-1',
            'test_normal_acinus_or_duct-0',
            'test_roi-0',
        }

        # deleting existing annotations in target slide (if any)
        delete_annotations_in_slide(girderClient, sampleSlideId)

        # post annotations to slide -- make sure it posts without errors
        resp = girderClient.post(
            '/annotation?itemId=' + sampleSlideId,
            json=annotation_docs[0])
        assert 'annotation' in resp.keys()
